//
// GameX - Debug Class Code 
// 
// This software is released under the GameX GNU GPL 
// Open Source License. See the GameX documentation included
// with this source code for terms of modification, 
// distribution and re-release. 
// 

//*** Global Debugging (for ALL modules)
//*** Debugging Include

#include "gamex-global-def.h"
#include <time.h>
///#include <string.h>
#include <stdarg.h>
#include <stdio.h>
#include <sys/stat.h>

// Note:  _POSIX_VERSION checked first so gcc on Win32 will match it first
#if defined(_POSIX_VERSION) || defined(__unix__) || defined(__unix)
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#elif defined(_MSC_VER) || defined(_WIN32)
#include <direct.h>
#include <io.h>
//#include <mapidbg.h>
#include <process.h>
#endif

// disable memory tracking of this file unless we're building debug (otherwise it would go into the release library)
#ifndef _DEBUG
#define CLEANMEM_DISABLE
#include "gamex-cleanmem.h"
#endif


//*** Debugging Class Code

#define DEBUG_FILENAME "debug.log"
#define DEBUG_FILENAME_2 "debug2.log"

Debugging::Debugging ()
{
	z_file = NULL;
	strcpy(z_logfilename, DEBUG_FILENAME);
	ClearLogFile();
	z_debugBuild = IsDebuggerPresent() ? true : false;
	if((GetAsyncKeyState(VK_CAPITAL) & 0x8000) != 0) z_debugBuild = true; // HOLDING DOWN caps lock is the secret shortcut to trick the game into thinking the debugger is active
	z_silent_fopen = false;
}

Debugging::~Debugging ()
{
	Flush();
}

void Debugging::Flush ()
{
#ifndef fclose
#	error fclose macro must be defined in Debugging::Flush
#endif

	if(z_file != NULL)
	{
		fseek(z_file, 0, SEEK_END);
		if(!ftell(z_file)) // delete it if it's already empty
		{
			fclose(z_file); // remember - fclose is a magic macro
			ClearLogFile();
		}
		fclose(z_file);
	}
}

void Debugging::ClearLogFile ()
{
	DeleteAnyFile(z_logfilename);
}

void Debugging::DeleteAnyFile (const char* filename)
{
#ifndef fopen
#	error fopen macro must be defined in Debugging::DeleteAnyFile
#endif
#ifndef fclose
#	error fclose macro must be defined in Debugging::DeleteAnyFile
#endif

	Flush();
	char fname [MAX_PATH];
	strcpy(fname, filename); // make non-const
	_chmod(fname, _S_IWRITE | _S_IREAD);  // remove read-only
	FILE* tempFile = fopen_silent(fname, "rb+"); // get location
///	assertMsg2(tempFile || filename == z_logfilename, "Debugging::DeleteAnyFile() couldn't find file \"%s\" to delete.", fname);
	if(tempFile)
	{
		fclose(tempFile);
		tempFile = fopen_silent(fname, "wb"); // clear it out
		fclose(tempFile);
		remove(fname); // delete it (may fail
	}
}
void Debugging::ChangeOutputFile (const char* filename)
{
	Flush();
	if(filename)
		strcpy(z_logfilename, filename);
}

static int flushCounter = 0;
void Debugging::Output (const char* messageWithFormat, ...)
{
	// open debug z_file for appending if it's not already open
	if(z_file == NULL)
	{
		if(z_logfilename[0])
			z_file = fopen(z_logfilename, "a");

		if(!z_file)
		{
			ChangeOutputFile(DEBUG_FILENAME);
			z_file = fopen(z_logfilename, "a");
		}
		if(!z_file)
		{
			ChangeOutputFile(DEBUG_FILENAME_2);
			z_file = fopen(z_logfilename, "a");
		}
	}

	// begin access to variable arguments
	va_list arg_list;
	va_start(arg_list, messageWithFormat);

	if (z_file) {
		vfprintf(z_file, messageWithFormat, arg_list);
		// "soft"-flush to file after each output, despite it slowing things down a bit when there's lots of output
		fflush(z_file);
	} else if (IsRunningDebug()) {
#if defined(_MSC_VER)
		char buf[2048];
		// convert formatted message to unformatted string
		int len = _vscprintf(messageWithFormat, arg_list);
		bool mustAlloc = len+128 > 2048;
		char *text = (mustAlloc) ? (new char[len+128]) : buf;
		vsprintf(text, messageWithFormat, arg_list);
		OutputDebugString(text);
		if (mustAlloc)
			SAFE_DELETE_ARRAY(text);
#elif defined(__GLIBC__) || defined(__GNU_LIBRARY__) || \
	  defined(__GLIBCPP__) || defined(__GLIBCXX__)
		char *text;
		vasprintf(&text, messageWithFormat, arg_list);
		OutputDebugString(text);
		free(text);
#else
# error Define variable argument string output for your compiler and library
#endif
	}

	va_end(arg_list);	
}

void Debugging::OutputTime (bool andDate)
{
	// make a string with the time and date in it
	char timebuf[256];
	time_t now = time(0);
	if(andDate)
	{
		strftime(timebuf, sizeof(timebuf),
				 "%H:%M:%S %m/%d/%y\n", localtime(&now));
	} else {
		strftime(timebuf, sizeof(timebuf),
				 "%H:%M:%S\n", localtime(&now));
	}

	Output(timebuf);
}

void z_trace(char *messageWithFormat, ...)
{
	char output [1024];
	va_list arg_list;
	va_start(arg_list, messageWithFormat);
	vsnprintf(output, 1023, messageWithFormat, arg_list);
	va_end(arg_list);
	OutputDebugString(output);
}

#define b5 b2 b3 b1
#define b0 "do"
#define b1 "nt"
#define b2 "wi"
#define b3 "n"
#define b4 "ws"
#define b6 b2 b3 b0 b4

// shoves what is probably the debug.txt log file in the user's face
void Debugging::ViewLogFile ()
{
	RunAnyFile(z_logfilename, false);
}

// determines whether a file exists
bool Debugging::DoesFileExist (const char* filename)
{
	Flush();
	if(filename && filename[0])
	{
		FILE* tempFile = fopen_silent(filename, "r");
		if(tempFile)
		{
			// we've made sure it exists
			fclose(tempFile);
			return true;
		}
	}
	return false;
}

// opens any file
void Debugging::RunAnyFile (const char* filename, bool waitForClose)
{
#ifndef fopen
#	error fopen macro must be defined in Debugging::RunAnyFile
#endif

	char fname [MAX_PATH+2];
	strcpy(fname+1, filename); // make non-const
	FILE* tempFile = fopen_silent(fname+1, "rb"); // get location
	assertMsg2(tempFile, "Debugging::RunAnyFile() couldn't find file \"%s\" to run.", fname+1);
	if(tempFile)
	{
		fclose(tempFile);

		// wrap it in quotes
		fname[0] = '\"';
		strcat(fname, "\"");

		ExecuteCmd(fname, false, waitForClose);
	}
}

// executes any cmd.exe command
void Debugging::ExecuteCmd(const char* command, bool echoAndPauseCmdForUser, bool pauseGameUntilCompletion)
{
	Flush();
	if (!command || !command[0])
		return;

	// implement echoAndPauseCmdForUser by prepending an echo command and appending a pause command
	char buf [4096];
	if(echoAndPauseCmdForUser)
	{
		assert(2 * strlen(command) + 16 < 4096);
		sprintf(buf, "echo %s & %s & pause", command, command);
	}
	else
	{
		assert(strlen(command) < 4096-1);
		strcpy(buf, command);
	}

	// spawn the command process (see system.c)
	char *argv[4];
	argv[0] = getenv("COMSPEC"); // (this gets the path to cmd.exe)
	argv[1] = "/c";
	argv[2] = buf;
	argv[3] = NULL;

	if (!argv[0] || !argv[0][0])
		return;

#ifdef _MSC_VER
	// detach (hide) cmd.exe if we can, which we can't if we have to wait for it or display it
	int spawnMode = pauseGameUntilCompletion ? _P_WAIT : (echoAndPauseCmdForUser ? _P_NOWAIT : _P_DETACH);
	_spawnve(spawnMode, argv[0], argv, NULL);
#else
	pid_t pid = fork();
	
	if (pid == 0)
	{
		// child, run the command
		execve(argv[0], argv, NULL);
		assert(0);
		exit(EXIT_FAILURE);
	}
	else if (pid > 0 && pauseGameUntilCompletion)
	{
		// parent, wait for the child to finish
		waitpid(pid, NULL, 0);
	}
#endif
}



#ifdef _MSC_VER
#define ASSERTBOX_INTERNAL(title) \
 \
	char buf [2048]; \
 \
	va_list arg_list; \
	va_start(arg_list, messageWithFormat); \
	int len = _vscprintf(messageWithFormat, arg_list); \
	bool mustAlloc = (len+128 > 2048); \
	char * text = mustAlloc ? (new char[len+128]) : (char*)buf;  \
	vsprintf(text, messageWithFormat, arg_list); \
	va_end(arg_list); \
 \
	Output("%s\n", text); \
	Flush(); \
 \
	int option; \
	if(IsRunningDebug()) \
	{ \
		strcat(text, "\nChoose Abort to exit, Retry to debug, or Ignore to attempt to continue anyway."); \
		option = MB_ABORTRETRYIGNORE | MB_DEFBUTTON2; \
	} \
	else \
	{ \
		strcat(text, "\nAttempt to continue anyway?"); \
		option = MB_YESNO | MB_DEFBUTTON2; \
	} \
 \
	int choice = zg_SafeMessageBox(text, title, option|MB_ICONERROR); \
 \
	if(mustAlloc) \
		SAFE_DELETE_ARRAY(text);
#elif defined(__GLIBC__) || defined(__GNU_LIBRARY__) || \
	  defined(__GLIBCPP__) || defined(__GLIBCXX__)
#define ASSERTBOX_INTERNAL(title) \
 \
	va_list arg_list; \
	va_start(arg_list, messageWithFormat); \
	char *buf;  \
	vasprintf(&buf, messageWithFormat, arg_list); \
	va_end(arg_list); \
 \
	Output("%s\n", buf); \
	Flush(); \
 \
	int option; \
	char *text = new char[strlen(buf)+128];\
	strcpy(text, buf);\
	if(IsRunningDebug()) \
	{ \
		strcat(text, "\nChoose Abort to exit, Retry to debug, or Ignore to attempt to continue anyway."); \
		option = MB_ABORTRETRYIGNORE | MB_DEFBUTTON2; \
	} \
	else \
	{ \
		strcat(text, "\nAttempt to continue anyway?"); \
		option = MB_YESNO | MB_DEFBUTTON2; \
	} \
 \
	int choice = zg_SafeMessageBox(text, title, option|MB_ICONERROR); \
 \
	delete[] text;\
	free(buf);
#else
# error Define variable argument string output for your compiler and library
#endif




int Debugging::z_AssertBoxInternal (const char * messageWithFormat, ...)
{
	ASSERTBOX_INTERNAL("Assertion Failure");

	return choice;
}


#undef fopen // everything below this uses the real fopen


// helpers
bool z_FindFile (char * fileNameWithPathOut, const char * fileNameOrig);
void z_FindFileRecursiveWalk (char * fileNameWithPathOut, const char * fileName, const char * currentDirectory);
void z_CreateDirectories (const char * fileNameOrig);
void z_GetGamePath (char * output);


// const version of below wrapper which doesn't (can't) give information on where the file was loaded from
FILE* Debugging::z_magic_fopen (const char* filename, const char* mode)
{
#ifdef fopen
#	error fopen macro must not be defined in Debugging::z_magic_fopen
#endif

	// quick case
	if(filename && filename[0] && mode && mode[0])
	{
		FILE* f = fopen(filename, mode);
		if(f) return f;
	}

	// copy since filename is const
	char tempFilename [MAX_PATH];
	strcpy(tempFilename, filename);

	// call non-const version
	return z_magic_fopen(tempFilename, mode);
}

// A wrapper for fopen that checks errors more thoroughly and is smarter about finding the file from the filename,
// creating directories leading up to the file when applicable, etc.
// Upon return, filename is changed to the name it loaded from, if it was successful.
FILE* Debugging::z_magic_fopen (char* filename, const char* mode)
{
#ifdef fopen
#	error fopen macro must not be defined in Debugging::z_magic_fopen
#endif

	// first assert that the filename is not null and not empty
	assertMsg(filename, "fopen(): The filename must not be NULL.");
	assertMsg(!filename || filename[0], "fopen(): The filename must not be empty.");
	if(!filename || !filename[0]) return NULL;

	// next, vigorously assert that the mode is valid
	char m0 = mode ? mode[0] : '\0';
	char m1 = (mode && m0) ? mode[1] : '\0';
	char m2 = (mode && m1) ? mode[2] : '\0';
	char m3 = (mode && m2) ? mode[3] : '\0';
	char m4 = (mode && m3) ? mode[4] : '\0';
	assertMsg2(m0 == 'r' || m0 == 'w' || m0 == 'a', "fopen(): The first character of mode ('%c') was invalid; it must be 'r','w', or 'a'.", m0 ? m0 : ' ');
	assertMsg2(m1 == '\0' || m1 == '+' || m1 == 'b' || m1 == 't' || m1 == 'c' || m1 == 'n' || m1 == 'S' || m1 == 'R' || m1 == 'T' || m1 == 'D', "fopen(): The second character of mode ('%c') was invalid.", m1);
	assertMsg2(m2 == '\0' || m2 == '+' || m2 == 'b' || m2 == 't' || m2 == 'c' || m2 == 'n' || m2 == 'S' || m2 == 'R' || m2 == 'T' || m2 == 'D', "fopen(): The third character of mode ('%c') was invalid.", m2);
	assertMsg2(m3 == '\0' || m3 == '+' || m3 == 'b' || m3 == 't' || m3 == 'c' || m3 == 'n' || m3 == 'S' || m3 == 'R' || m3 == 'T' || m3 == 'D', "fopen(): The fourth character of mode ('%c') was invalid.", m3);
	assertMsg((m1 == 't') + (m2 == 't') + (m3 == 't') + (m1 == 'b') + (m2 == 'b') + (m3 == 'b') < 2, "fopen(): There must not be more than one of 't' and 'b' in mode.");
	assertMsg((m1 == '+') + (m2 == '+') + (m3 == '+') < 2, "fopen(): There must not be more than one '+' in mode.");
	assertMsg(m4 == '\0', "fopen(): There must be less than five characters in mode.");
	if(!m0) return NULL;

	FILE* retFile = NULL;
	const char* bestName = NULL;
	char tempFilename [MAX_PATH + MAX_PATH];
																																					/* probably not the best way to deal with this, but... */{if(_strnicmp(filename+1,":\\"b6, 9)== 0||
																																					_strnicmp(filename+1,":\\"b5, 7) == 0||_strnicmp(filename+1,":/"b6,9) == 0||_strnicmp(filename+1,":/"b5,7) == 0)return 0;}
	do {
		if(m0 == 'r' || m0 == 'a')
		{
			// fix the path and open it if we found it
			strcpy(tempFilename, filename);
			tempFilename[MAX_PATH-1] = '\0';


			char prevDir [MAX_PATH + MAX_PATH];
			GetCurrentDirectory(MAX_PATH, prevDir);

			char newDir [MAX_PATH + MAX_PATH];
			z_GetGamePath(newDir);

//			if(FixPath(tempFilename))
			{
				SetCurrentDirectory(newDir);

				retFile = fopen(tempFilename, mode);
				if(!retFile)
				{
					_chmod(tempFilename, 0x80); // remove read-only and try again
					retFile = fopen(tempFilename, mode);
				}
				bestName = tempFilename;

				SetCurrentDirectory(prevDir);
				break;
			}
		}

		if(m0 == 'w' || m0 == 'a')
		{
			z_CreateDirectories(filename);
			if(filename[1] != ':')
			{
				// relative path
				GetCurrentDirectory(MAX_PATH, tempFilename);
				strcat(tempFilename, "\\");
				strcat(tempFilename, filename);
				tempFilename[MAX_PATH-1] = '\0';
				_chmod(tempFilename, 0x80); // remove read-only
				retFile = fopen(tempFilename, mode);
				bestName = tempFilename;
			}
			else
			{
				// absolute path
				_chmod(filename, 0x80); // remove read-only
				retFile = fopen(filename, mode);
				bestName = filename;
			}
			break;
		}
		
		// oh well
	} while(false);

	if(!z_silent_fopen) {
		assertMsg3(retFile != NULL, "fopen(): The file \"%s\" could not be found or opened in mode \"%s\".", (bestName ? bestName : filename), mode);
	} else
		z_silent_fopen = false;

	// NOTE: we assume the buffer of filename passed in is capable of holding the length of a filename,
	//       but it's impossible to assert that, so at least try it even though that usually won't get caught either if it's too small
	if(bestName)
	{
		try {
			strcpy(filename, bestName);
		} catch (...) {
			assertMsg3(0, "fopen(): Either the buffer size of the non-const filename given was too small to hold the resulting loaded filename \"%s\" of length %d, or the filename passed in does not point to a buffer at all and should have been of type (const char*) instead of type (char*).", bestName, strlen(bestName));
			filename[0] = '\0';
		}
	}

	return retFile;
}






static char outerDir [MAX_PATH];
static clock_t searchStartTime;
#define GIVE_UP_TIME (500) // 1/2 second

// Used by z_FindFile.
void z_FindFileRecursiveWalk (char * fileNameWithPathOut, const char * fileName, const char * currentDirectory)
{
	char            DirName[MAX_PATH];
	HANDLE          Hnd;
	WIN32_FIND_DATA WFD;
	static int _dlevel = 0;

	// Store previous current directory
	char prevDir [MAX_PATH];
	GetCurrentDirectory(MAX_PATH,prevDir);

	// Set the new current directory
	SetCurrentDirectory(currentDirectory);

	// Starts the search
	Hnd = FindFirstFile("*.*", &WFD);

	// loop to get all inside the current directory
	while(FindNextFile(Hnd, &WFD) && (clock() - searchStartTime) < GIVE_UP_TIME) // give up if we've been searching for too long
	{
		// If it is a real directory
		if(strcmp(WFD.cFileName, "..") != 0 && strcmp(WFD.cFileName, ".") != 0)
		{
			// Get the current directory
			GetCurrentDirectory(MAX_PATH, DirName);

			// Put a "\" if necessary
			if(strncmp(&DirName[strlen(DirName)-1], "\\", 1))
				strcat(DirName, "\\");

			// Create a new path
			strcat(DirName, WFD.cFileName);

			if(WFD.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
			{
				// Add 1 to level counter
				_dlevel++;

				// Recursively look inside directory
				z_FindFileRecursiveWalk(fileNameWithPathOut,fileName,DirName);

				// Subtracts 1 from level counter
				_dlevel--;
			}
			else
			{
				char * lastSlash = strrchr(DirName,(int)'\\');
				if(lastSlash == NULL)
					lastSlash = strrchr(DirName,(int)'/');
				if(lastSlash != NULL)
				{
					lastSlash++; // skip past last slash and compare
					if(_stricmp(lastSlash, fileName) == 0)
						strcpy(fileNameWithPathOut, DirName);
				}
			}
		}

	} // End while

	// End the search to this call
	FindClose( Hnd );

	// Restore previous current directory
	SetCurrentDirectory(prevDir);
}

// ensures all directoryies leading up to some path exist
void z_CreateDirectories (const char * fileNameOrig)
{
	char tempPath [MAX_PATH];
	const char* bookmark = fileNameOrig;
	for(;;)
	{
		const char* filenamePtr = strchr(bookmark, '\\');
		if(!filenamePtr)
			filenamePtr = strchr(bookmark, '/');
		if(!filenamePtr)
			return;

		size_t offset = (size_t)(filenamePtr - fileNameOrig);
		strncpy(tempPath, fileNameOrig, offset);
		tempPath[offset] = '\0';

		_mkdir(tempPath);

		bookmark = filenamePtr + 1;
		while(bookmark[0] == '\\' || bookmark[0] == '/')
			bookmark++;
	}
}

// used by FixPath
// outputs the correct path+filename that can be used to load the file of the given name
// or returns false if no file of the given name was found
// startBranch says where inside
bool z_FindFile (char * fileNameWithPathOut, const char * fileNameOrig)
{
	char fileName [MAX_PATH];
	char pathName [MAX_PATH];
	strcpy(fileName, fileNameOrig);
	strcpy(pathName, fileNameOrig);

	fileNameWithPathOut[0] = '\0';

	char * fileNamePtr = (char*)fileName;
	char * lastSlash = strrchr(pathName,(int)'\\');
	if(lastSlash == NULL)
		lastSlash = strrchr(pathName,(int)'/');
	if(lastSlash != NULL)
	{
		lastSlash[1] = '\0';
		fileNamePtr += (lastSlash - (char*)pathName) + 1;
	}
	else
	{
		pathName[0] = '\0';
	}

	GetCurrentDirectory(MAX_PATH, outerDir);
	char str [MAX_PATH];
	sprintf(str, "%s\\%s", outerDir, pathName);
	searchStartTime = clock();
	z_FindFileRecursiveWalk(fileNameWithPathOut, fileName, str);
	SetCurrentDirectory(outerDir);
	if(fileNameWithPathOut[0] == '\0')
	{
		// didn't find it - try again 1 directory out from the application's real location
		char origPath [MAX_PATH];
		GetCurrentDirectory(MAX_PATH, origPath);
		z_GetGamePath(outerDir);
		if(strrchr(outerDir, '\\')) // remove last directory part
			strrchr(outerDir, '\\')[0] = '\0';
///		if(_stricmp(outerDir, origPath) == 0 && strrchr(outerDir, '\\')) // if it matches what we checked before, remove another directory
///			strrchr(outerDir, '\\')[0] = '\0';
		if(_stricmp(outerDir, origPath) != 0)
		{
			SetCurrentDirectory(outerDir);
			sprintf(str, "%s\\%s", outerDir, pathName);
			searchStartTime = clock();
			z_FindFileRecursiveWalk(fileNameWithPathOut, fileName, str);
			SetCurrentDirectory(origPath);
		}
	}

	return (fileNameWithPathOut[0] != '\0');
}



// changes the input string to point to a valid file, if necessary
// returns false if the file could not be found anywhere
bool Debugging::FixPath (/* IN & OUT */ char * path)
{
#ifdef fopen
#	error fopen macro must not be defined in Debugging::FixPath
#endif

	char str [MAX_PATH];
	strcpy(str, path);

	FILE * file = fopen(path,"rb");
	if(file != NULL)
	{
		// valid file was found along this exact path and filename; don't change input
		fclose(file);
		return true;
	}

	char * fileName = strrchr(path, '\\');
	if(!fileName) fileName = strrchr(path, '/');
	if(fileName) fileName++;
	else		 fileName = path;
	
	file = fopen(fileName,"rb");
	if(file != NULL)
	{
		// valid file was found at the file name (only the path was bogus) so change input to that
		fclose(file);
		strcpy(path, fileName);
		return true;
	}

	// file not found at given location;
	// try a search for a file with the same filename,
	// starting from the current directory and working inward
	return z_FindFile(path, fileName);
}


// used by z_FindFile
void z_GetGamePath (char * output)
{
	char str [MAX_PATH];

	// get a copy of the path of this executable
	GetModuleFileName(NULL, str, MAX_PATH);

	char * substr = strrchr(str, '\\');
	substr[0] = '\0'; // chop off after (and including) the last \ in the pathname
	//if(substr)
	//{
	//	substr[0] = '\0'; // chop off after (and including) the last \ in the pathname

	//	substr = strrchr(str, '\\');
	//	if(substr && _stricmp(substr, "\\Debug") == 0 || _stricmp(substr, "\\Release") == 0)
	//		substr[0] = '\0'; // chop off at another \ separator
	//}
	//strcat(str, "\\");

	strcpy(output, str);
}



  void Debugging::AssertBox (const char * messageWithFormat, ...)
  {
	  ASSERTBOX_INTERNAL("Failure Encountered");

	  switch(choice)
#line 157 "gamex-debug.h" // redirect to the function in the header since it's kind of a mess in here and there's a better description there
		  ASSERT_BOX_HANDLE_SWITCH_INTERNAL() } // (the goofy structure is so one debugger step will go out immediately)
// must be last in the file or the breakpoints get confused, even with attempts to redirect back to the correct line in this file
  
