//
// GameX - GameXClass Class Code
// NOTE: Not all GameXClass functions are implemented in this file
// 
// Copyright (C) 2002 Rama C. Hoetzlein
// 
// This software is released under the GameX GNU GPL 
// Open Source License. See the GameX documentation included
// with this source code for terms of modification, 
// distribution and re-release. 
// 
// this is a version of GameX that has been modified (by Justin Pease)



// DirectDraw
#ifndef DIRECTDRAW_VERSION
#define DIRECTDRAW_VERSION 0x0700 // specify version 7.0 of DirectDraw
#include <ddraw.h>
#endif

// Direct3D
#ifndef DIRECT3D_VERSION
#define DIRECT3D_VERSION 0x0700 // specify version 7.0 of Direct3D -- the last version that's compatible with DirectDraw
#include <d3d.h>				// note: for newer versions of D3D we would need to 
//#include <d3dtypes.h>			//		 include different files, like d3d8.h
//#include "d3dx.h"
#endif

#include "gamex-win-dx.h"
#include "gamex-camera.h"
#include "gamex-image.h"
#include "gamex-text.h"


///DWORD triangleListIndex [D3DMAXNUMVERTICES*6/4];


// disable memory tracking of this file unless we're building debug (otherwise it would go into the release library)
#ifndef _DEBUG
#define CLEANMEM_DISABLE
#include "gamex-cleanmem.h"
#endif

#define PUSH_VBUF()
#define POP_VBUF()
//#define PUSH_VBUF() const bool vBufferWasEnabled = z_d3d_vbufferenabled; if(!vBufferWasEnabled) GameX.EnableVertexBufferBatching()
//#define POP_VBUF() if(!vBufferWasEnabled) GameX.DisableVertexBufferBatching()
 

bool GameXClass::z_LockBackBuffer ()
{
	if(z_drawLocked) return true;
	
	// since this means a switch from D3D to DD drawing,
	// we have to interrupt batch rendering here
	if(z_draw_render_batch_open)
		z_EndRenderBatch();

	DDSURFACEDESC2		ddsd;

	ddsd.dwSize = sizeof(ddsd) ;

	HRESULT status = z_dd_back_buf->Lock(NULL, &ddsd, DDLOCK_WAIT | DDLOCK_SURFACEMEMORYPTR, NULL);	// lock drawing surface
	if(FAILED(status))
	{
		z_DirectDrawError(status);
		return false;
	}

	z_win_screen = (char *) ddsd.lpSurface;

	z_drawLocked = true;

	return true;
}

void GameXClass::z_UnlockBackBuffer ()
{
	if(z_drawLocked)
	{
		z_dd_back_buf->Unlock(NULL);
		z_drawLocked = false;
	}
} 

void GameXClass::SetSortingMode (SceneSortMode sortMode)
{
	z_draw_sort_mode = sortMode;

	if(!z_d3d_zbuffer_enabled) // resort to depth sorting if Z-buffering isn't working
		if((z_draw_sort_mode & SCENESORTZR) && (z_draw_sort_mode & SCENESORTZW))
			z_draw_sort_mode = SCENESORTD;

	if(z_draw_sort_mode & SCENESORTD)
		z_draw_3D_depth_sort = 1;
	else
		z_draw_3D_depth_sort = 0;

	if(z_draw_sort_mode & SCENESORTZR)
		z_d3d_device->SetRenderState(D3DRENDERSTATE_ZENABLE,TRUE);
	else
		z_d3d_device->SetRenderState(D3DRENDERSTATE_ZENABLE,FALSE);

	if(z_draw_sort_mode & SCENESORTZW)
		z_d3d_device->SetRenderState(D3DRENDERSTATE_ZWRITEENABLE,TRUE);
	else
		z_d3d_device->SetRenderState(D3DRENDERSTATE_ZWRITEENABLE,FALSE);
}

static bool needReClearZ = false;

void GameXClass::Begin3DScene ()
{
	SetSortingMode(SCENESORTZR|SCENESORTZW|SCENESORTD);

	// modify clipping boundaries
	{
		int x1, y1, x2, y2;
		z_draw_state->cam->GetWindow(x1, y1, x2, y2);

		// for each camera region boundary that's inside the draw dest boundaries,
		// replace the boundary value with that of the camera's

		if(x1 > z_draw_state->dst_bounds.left
		&& x1 < z_draw_state->dst_bounds.right)
			z_draw_state->dst_bounds.left = x1;

		if(y1 > z_draw_state->dst_bounds.top
		&& y1 < z_draw_state->dst_bounds.bottom)
			z_draw_state->dst_bounds.top = y1;

		if(x2 > z_draw_state->dst_bounds.left
		&& x2 < z_draw_state->dst_bounds.right)
			z_draw_state->dst_bounds.right = x2;

		if(y2 > z_draw_state->dst_bounds.top
		&& y2 < z_draw_state->dst_bounds.bottom)
			z_draw_state->dst_bounds.bottom = y2;
	}

	if(needReClearZ)
	{
		if(z_d3d_zbuffer_enabled)
			z_d3d_device->Clear(0,NULL,D3DCLEAR_ZBUFFER,0,1.0f,0); // clear the z-buffer
		needReClearZ = false;
	}
}


void GameXClass::End3DScene()
{
	if(z_draw_sort_mode & SCENESORTZW)
		z_d3d_device->SetRenderState(D3DRENDERSTATE_ZWRITEENABLE,FALSE);

///	if(z_draw_sort_mode & SCENESORTD)
	if(!z_draw_state_list.empty())
	{
		// remember the state
		GameXDrawState * initial = z_draw_state;

		// sort all the transparent drawing commands:
		z_draw_state_list.sort();

		z_draw_3D_depth_sort = 2;

		// draw all the commands
		for(list<GameXDrawState>::iterator iter = z_draw_state_list.begin() ; iter != z_draw_state_list.end() ; iter++)
		{
			z_draw_state = &(*iter);

			if(z_draw_state->flags & DRAWOP_ALPHAINTERSECT
			&& z_draw_state->flags & DRAW_PLAIN
			&& z_draw_state->src != VIEWPORT && z_draw_state->src != NULL
			&& z_draw_state->src->z_usage & LOAD_ALPHA
			&& z_draw_sort_mode & SCENESORTZW && z_draw_state->dst == VIEWPORT)
			{
				// if it meets certain conditions and DRAWOP_ALPHAINTERSECT is enabled,
				// draw it twice to allow the alpha image to intersect in 3D:
				z_d3d_device->SetRenderState(D3DRENDERSTATE_ALPHATESTENABLE, TRUE);
				z_d3d_device->SetRenderState(D3DRENDERSTATE_ZWRITEENABLE,TRUE);
				bool keep = (z_draw_state->flags & DRAWOP_KEEPSTATES) ? true : false;
				if(!keep) z_draw_state->flags |= DRAWOP_KEEPSTATES;
				z_InternalDrawMaster(0);
				z_d3d_device->SetRenderState(D3DRENDERSTATE_ALPHATESTENABLE, FALSE);
				z_d3d_device->SetRenderState(D3DRENDERSTATE_ZWRITEENABLE,FALSE);
				if(!keep) z_draw_state->flags ^= DRAWOP_KEEPSTATES;
			}

			z_InternalDrawMaster(0);
		}

		// clear the list for next time
		z_draw_state_list.clear();

		// put the drawing state back how it was
		z_draw_state = initial;

		z_draw_3D_depth_sort = 0;
	}

	if(z_draw_sort_mode & SCENESORTZR)
	{
		z_d3d_device->SetRenderState(D3DRENDERSTATE_ZENABLE,FALSE); 

		if(z_d3d_device != NULL)
		{
			if(z_draw_sort_mode & SCENESORTS)
			{
				if(!z_win_alternator)
					z_d3d_device->SetRenderState(D3DRENDERSTATE_ZFUNC, D3DCMP_GREATEREQUAL);
				else
					z_d3d_device->SetRenderState(D3DRENDERSTATE_ZFUNC, D3DCMP_LESSEQUAL);
			}
			else
			{
				needReClearZ = true;
//				z_d3d_device->Clear(0,NULL,D3DCLEAR_ZBUFFER,0,1.0f,0); // clear the z-buffer
			}
		}
	}

	z_draw_sort_mode = 0;

	// revert clipping boundaries
	SetDrawDestination(GetDrawDestination());
}

int totalStrips = 0;
int totalAlone = 0;


void GameXClass::z_EmptyVBuffer ()
{
	if(z_d3d_vbufferlocked && z_d3d_vbuffer != NULL)
	{
///		z_d3d_device->SetTexture(0,NULL);

		IDirect3DVertexBuffer7 & v = *z_d3d_vbuffer;
		v.Unlock();
		z_d3d_vbuffermem = NULL;

		if(z_d3d_vbufferfilled)
		{
	//		z_d3d_device->DrawIndexedPrimitiveVB(D3DPT_TRIANGLELIST, z_d3d_vbuffer, 0, z_d3d_vbufferfilled, triangleListIndex, z_d3d_vbufferfilled*6/4, 0);

	/*
			for(int i = 0 ; i < z_d3d_vbufferfilled ; i += 4)
			{
				z_d3d_device->DrawPrimitiveVB(D3DPT_TRIANGLESTRIP, z_d3d_vbuffer, i, 4, 0);
			}
	*/
	//		z_d3d_device->DrawPrimitiveVB(D3DPT_TRIANGLELIST, z_d3d_vbuffer, 0, z_d3d_vbufferfilled, 0);

			z_d3d_device->DrawPrimitiveVB(D3DPT_TRIANGLESTRIP, z_d3d_vbuffer, 0, z_d3d_vbufferfilled, 0);
			totalStrips++;

			//static long long total = 0;
			//static long long totalTimes = 0;
			//total += z_d3d_vbufferfilled;
			//totalTimes++;
			//debug.Output("%d\n", total / totalTimes);

			z_d3d_vbufferfilled = 0;
		}

		z_d3d_vbufferlocked = false;

	}
}

void GameXClass::z_LockVBuffer ()
{
	if(z_d3d_vbufferenabled && !z_d3d_vbufferlocked && z_d3d_vbuffer != NULL)
	{
		IDirect3DVertexBuffer7 & v = *z_d3d_vbuffer;
		DWORD flags = DDLOCK_DISCARDCONTENTS | DDLOCK_WAIT | DDLOCK_WRITEONLY | DDLOCK_SURFACEMEMORYPTR;

		if(GameX.z_win_log_debug)
			flags |= DDLOCK_NOSYSLOCK; // prevent possible win16 mutex debugger freeze

///		z_d3d_vbuffermem = NULL;
//		z_d3d_vbufferstride = sizeof(DWORD) // D3DFVF_DIFFUSE
//						  + sizeof(float) * 2 * 1 // D3DFVF_TEX1
//						  + sizeof(float) * 4 // D3DFVF_XYZRHW
//						  ;
//		int stridecheck = sizeof(FVFTLVERTEX);
		v.Lock(flags, (void**)&z_d3d_vbuffermem, NULL);

		z_d3d_vbufferlocked = true;
	}
}

// returns whether the drawing should be skipped for now
bool GameXClass::z_GetScreenVertices (const ImageX * source, bool immediate)
{
	z_draw_state->src = source;

	if(!z_GetScreenVertices(z_draw_state->x,z_draw_state->y,z_draw_state->z))
		return true;

	if(z_draw_3D_depth_sort == 1) // if possibly storing sorted states
	{
		z_draw_state->zcomp = z_draw_state->z[0]+z_draw_state->z[1]+z_draw_state->z[2]+z_draw_state->z[3];

		if(!immediate)
		{
			z_draw_state_list.push_back(*z_draw_state);
			return true;
		}
	}

	return false;
}


// convert draw modes to the closest matching combiner operator
// NOTE: this is COMPLETELY UNRELATED to drawing an image in a certain draw mode like DRAW_ADD,
// this is only used to set how color shading is done, i.e. how the color set by SetDrawShading is used.
// ALSO NOTE: this does not affect how the alpha component of the color is used.
void GameXClass::SetDrawShadingEffect(DrawFlags effect)
{
	switch(effect & DRAWMODESMASK)
	{
		case 0: z_draw_state->color_effect = D3DTOP_SELECTARG1; break; // if no draw mode given, disable shading
		case DRAW_PLAIN: z_draw_state->color_effect = D3DTOP_SELECTARG2; break; // plain -> the color takes over
		case DRAW_ADD: case DRAW_ADDSHARP: z_draw_state->color_effect = D3DTOP_ADD; break; // shade color is added to the image being drawn
		case DRAW_GHOST: z_draw_state->color_effect = D3DTOP_ADDSIGNED; break; // can lighten or darken, usually changes lightness instead of color, often faint
		case DRAW_INTENSIFY: case DRAW_ADDSOFT: z_draw_state->color_effect = D3DTOP_ADDSMOOTH; break; // similar to with DRAW_ADD
		case DRAW_SUBTRACT: case DRAW_BURN: z_draw_state->color_effect = D3DTOP_SUBTRACT; break; // shade color subtracted from image -- interesting effect
		case DRAW_INVERT: z_draw_state->color_effect = D3DTOP_DOTPRODUCT3; break; // not inversion, but it's interesting, albeit black-and-white
		default: case DRAW_MULTIPLY: z_draw_state->color_effect = D3DTOP_MODULATE; break; // shade color modulates image color -- normally desired effect
	}
}

int GameXClass::z_ClipRects (RECT & src_rect, RECT & dst_rect, const RECT & src_bounds, const RECT & dst_bounds)
{
	if(dst_rect.right >= dst_bounds.left && dst_rect.left <= dst_bounds.right) { // In-View Horizontally
		if(dst_rect.bottom >= dst_bounds.top && dst_rect.top <= dst_bounds.bottom) { // In-View Vertically 

			if(dst_rect.right-dst_rect.left == src_rect.right-src_rect.left
			&& dst_rect.bottom-dst_rect.top == src_rect.bottom-src_rect.top) { // If we don't need to scale

				if(dst_rect.left < dst_bounds.left)
				{
					src_rect.left += dst_bounds.left - dst_rect.left;
					dst_rect.left = dst_bounds.left;
				}
				if(dst_rect.top < dst_bounds.top)
				{
					src_rect.top += dst_bounds.top - dst_rect.top;
					dst_rect.top = dst_bounds.top;
				}
				if(dst_rect.right > dst_bounds.right)
				{
					src_rect.right += dst_bounds.right - dst_rect.right;
					dst_rect.right = dst_bounds.right;
				}
				if(dst_rect.bottom > dst_bounds.bottom)
				{
					src_rect.bottom += dst_bounds.bottom - dst_rect.bottom;
					dst_rect.bottom = dst_bounds.bottom;
				}
				// clip the source too... we could skip this is we were lazy, but then
				// giving an invalid source position would cause the game to crash
				if(src_rect.left < src_bounds.left)
				{
					dst_rect.left += src_bounds.left - src_rect.left;
					src_rect.left = src_bounds.left;
				}
				if(src_rect.top < src_bounds.top)
				{
					dst_rect.top += src_bounds.top - src_rect.top;
					src_rect.top = src_bounds.top;
				}
				if(src_rect.right > src_bounds.right)
				{
					dst_rect.right += src_bounds.right - src_rect.right;
					src_rect.right = src_bounds.right;
				}
				if(src_rect.bottom > src_bounds.bottom)
				{
					dst_rect.bottom += src_bounds.bottom - src_rect.bottom;
					src_rect.bottom = src_bounds.bottom;
				}
				if(dst_rect.left >= dst_rect.right || dst_rect.top >= dst_rect.bottom)
					return 0;
				return 1; // Image is in view and should be drawn as-is with the src_rect and dst_rect that we just (possibly) modified
			} else { // need to scale, and if we clip it we need to adjust the clipping amount to match
				
				// the clipping used here isn't perfect since it doesn't allow for pixels to be partially clipped,
				// but it's better than blindly assuming a 1:1 source:destination ratio

				if(dst_rect.left < dst_bounds.left)
				{
					src_rect.left += (dst_bounds.left - dst_rect.left) * (src_rect.right - src_rect.left) / (dst_rect.right - dst_rect.left);
					dst_rect.left = dst_bounds.left;
				}
				if(dst_rect.top < dst_bounds.top)
				{
					src_rect.top += (dst_bounds.top - dst_rect.top) * (src_rect.bottom - src_rect.top) / (dst_rect.bottom - dst_rect.top);
					dst_rect.top = dst_bounds.top;
				}
				if(dst_rect.right > dst_bounds.right)
				{
					src_rect.right += (dst_bounds.right - dst_rect.right) * (src_rect.right - src_rect.left) / (dst_rect.right - dst_rect.left);
					dst_rect.right = dst_bounds.right;
				}
				if(dst_rect.bottom > dst_bounds.bottom)
				{
					src_rect.bottom += (dst_bounds.bottom - dst_rect.bottom) * (src_rect.bottom - src_rect.top) / (dst_rect.bottom - dst_rect.top);
					dst_rect.bottom = dst_bounds.bottom;
				}
				if(src_rect.left < src_bounds.left)
				{
					dst_rect.left += (src_bounds.left - src_rect.left) * (dst_rect.right - dst_rect.left) / (src_rect.right - src_rect.left);
					src_rect.left = src_bounds.left;
				}
				if(src_rect.top < src_bounds.top)
				{
					dst_rect.top += (src_bounds.top - src_rect.top) * (dst_rect.bottom - dst_rect.top) / (src_rect.bottom - src_rect.top);
					src_rect.top = src_bounds.top;
				}
				if(src_rect.right > src_bounds.right)
				{
					dst_rect.right += (src_bounds.right - src_rect.right) * (dst_rect.right - dst_rect.left) / (src_rect.right - src_rect.left);
					src_rect.right = src_bounds.right;
				}
				if(src_rect.bottom > src_bounds.bottom)
				{
					dst_rect.bottom += (src_bounds.bottom - src_rect.bottom) * (dst_rect.bottom - dst_rect.top) / (src_rect.bottom - src_rect.top);
					src_rect.bottom = src_bounds.bottom;
				}
				if(dst_rect.left >= dst_rect.right || dst_rect.top >= dst_rect.bottom)
					return 0;
				return 2; // Image is in view and should be drawn SCALED with the src_rect and dst_rect that we just (possibly) modified
			}
		}
	}
	return 0; // Image is not in view at all and so shouldn't be drawn
}



void GameXClass::z_ClipTile (int x, int y, int xres, int yres, int &mx, int &my, int &offx, int &offy)
{
	if(!xres || !yres)
		return;

	if(x<0)
	{
		mx = ((int) (x / xres)) - 1;
		offx = (x % xres) + xres;
	}
	else
	{
		mx = ((int) (x / xres)) ;
		offx = x % xres;
	}
	if(y<0)
	{
		my = ((int) (y / yres)) - 1;		
		offy = (y % yres) + yres;
	}
	else
	{
		my = ((int) (y / yres)) ;		
		offy = y % yres;
	}	
}

// Params: left, top, right, bottom, red (out of 255), green (out of 255), blue (out of 255)
// (this is the toned-down but ultra-fast version of DrawRect)
void GameXClass::z_InternalDrawRectDD (int r, int g, int b, int x1, int y1, int x2, int y2)
{
	z_pushDrawing();
	if(x2 <= x1)
	{
		if(z_draw_state->flags & DRAWOP_NOCULL && x1 != x2)
		{
			int t = x2;
			x2 = x1;
			x1 = t;
		} else goto funcEnd;
	}
	if(y2 <= y1)
	{
		if(z_draw_state->flags & DRAWOP_NOCULL && y1 != y2)
		{
			int t = y2;
			y2 = y1;
			y1 = t;
		} else goto funcEnd;
	}

	{
		RECT dst_rect = {x1,y1,x2,y2};

		if(!z_dd_back_buf)
		{
			if(!z_win_hwnd)
				goto funcEnd;

			HDC hdc = GetDC(z_win_hwnd);

			if(!hdc)
				goto funcEnd;

			HRGN hrgn = CreateRectRgn(z_win_viewx1, z_win_viewy1, z_win_viewx2, z_win_viewy2);
			FillRgn(hdc, hrgn, (HBRUSH)GetStockObject(BLACK_BRUSH));

			ReleaseDC(z_win_hwnd, hdc);

			goto funcEnd;
		}

		z_UnlockBackBuffer() ; // unlock drawing surface if necessary

		DDBLTFX ddbltfx;
		ddbltfx.dwSize = sizeof(ddbltfx);
		ddbltfx.dwFillColor = (((r >> (8-GameX.z_win_maxr)) << z_win_shiftr) + ((g >> (8-GameX.z_win_maxg)) << z_win_shiftg) + ((b >> (8-GameX.z_win_maxb)) << z_win_shiftb));

		HRESULT status = z_dd_back_buf->Blt(&dst_rect, NULL, NULL, DDBLT_WAIT | DDBLT_COLORFILL, &ddbltfx);
		if(FAILED(status))
			z_DirectDrawError(status);
		else goto funcEnd;
	
		// if we failed, try it the old way:

		z_LockBackBuffer() ;

		unsigned short *start, *pixel;
		int i, j, clr;

		clr = ddbltfx.dwFillColor;
		start = (unsigned short *) z_win_screen;
		start += y1*z_win_pitch + x1;
		for(j = y1; j <= y2; j++)
		{
			pixel = start;
			for(i = x1; i <= x2; i++)
				*pixel++ = clr;
			start += z_win_pitch;
		}
	}

funcEnd:
	z_popDrawing();
}

// 2D drawing functions:
void GameXClass::DrawPoint	(const ColorX & clr, float x, float y)
{
	z_draw_state->colors = clr; 
	z_draw_state->vertices_3D = 0;
	z_draw_state->dst_quad2d = Quad2D((float)x,(float)y,DEFAULT2,DEFAULT2);
	z_InternalDrawMaster(NULL);
}
void GameXClass::DrawLine	(const ColorX & clr, float x1,float y1, float x2,float y2) 
{
	z_draw_state->colors = clr;
	z_draw_state->angle = 0.0f;
	z_draw_state->vertices_3D = -2;
	z_draw_state->dst_quad2d = Quad2D((float)x1,(float)y1,(float)x2,(float)y2);
	z_InternalDrawMaster(NULL);
}
void GameXClass::DrawRect	(const ColorX & clr, float xLeft, float yTop, float xRight, float yBottom)
{
	z_draw_state->colors = clr;
	z_draw_state->vertices_3D = 0;
	z_draw_state->dst_quad2d = Quad2D((float)xLeft,(float)yTop,(float)xRight,(float)yBottom);
	z_InternalDrawMaster(NULL);
}

void GameXClass::DrawPolygon (const ColorX & clr, float x1,float y1, float x2,float y2, float x3,float y3, float x4,float y4)
{
	z_draw_state->colors = clr;
	z_draw_state->vertices_3D = 0;
	z_draw_state->dst_quad2d = Quad2D((float)x1,(float)y1,(float)x2,(float)y2,(float)x3,(float)y3,(float)x4,(float)y4);
	if(!(z_draw_state->flags & DRAWOP_NOCULL))
		z_draw_state->flags |= DRAWOP_NOCULL;
	z_InternalDrawMaster(NULL);
}

// 2D image drawing functions:
void GameXClass::DrawImage (const ImageX * img, float x, float y)
{
	z_draw_state->vertices_3D = 0;
	z_draw_state->dst_quad2d = Quad2D(floorf(x),floorf(y),(float)DEFAULT2,(float)DEFAULT2);
	z_InternalDrawMaster(img);
}
void GameXClass::DrawImage (const ImageX * img, float x, float y, float angle, float scale) /* shorthand for calling SetDrawAngle, SetDrawScale before drawing DrawImage */
{
	SetDrawAngle(angle);
	SetDrawScale(scale);
	DrawImage(img,x,y);
}
void GameXClass::DrawLineImage (const ImageX * img, float x1,float y1, float x2,float y2, bool pixelUnitScale)
{
	z_draw_state->angle = 0.0f;
	z_draw_state->vertices_3D = pixelUnitScale ? -2 : -1;
	z_draw_state->dst_quad2d = Quad2D((float)x1,(float)y1,(float)x2,(float)y2);
	z_InternalDrawMaster(img);
}
void GameXClass::DrawTexturedRect (const ImageX * img, float xLeft, float yTop, float xRight, float yBottom)
{
	z_draw_state->src = img;
	z_draw_state->vertices_3D = 0;
	z_draw_state->dst_quad2d = Quad2D((float)xLeft,(float)yTop,(float)xRight,(float)yBottom);
	z_InternalDrawMaster(img);
}
void GameXClass::DrawTexturedPolygon (const ImageX * img, float x1,float y1, float x2,float y2,  float x3,float y3,  float x4,float y4)
{
	z_draw_state->src = img;
	z_draw_state->vertices_3D = 0;
	z_draw_state->dst_quad2d = Quad2D((float)x1,(float)y1,(float)x2,(float)y2,(float)x3,(float)y3,(float)x4,(float)y4);
	if(!(z_draw_state->flags & DRAWOP_NOCULL)) z_draw_state->flags |= DRAWOP_NOCULL; // culling of 2D polys would cause confusion
	z_InternalDrawMaster(img);
}

// 3D drawing functions:
void GameXClass::DrawPoint3D   (const ColorX & clr, const Vec3f & v)
{
	z_draw_state->colors = clr;
	z_draw_state->vertices_3D = 3;
	z_draw_state->dst_point3d = v;
	z_InternalDrawMaster(NULL);
}
void GameXClass::DrawLine3D	  (const ColorX & clr, const Vec3f & v1, const Vec3f & v2)
{
	z_draw_state->colors = clr;
	z_draw_state->vertices_3D = 5;
	z_draw_state->dst_quad3d = Quad3D(v1,v2,v1,v2);
	z_InternalDrawMaster(NULL);
}
void GameXClass::DrawPolygon3D (const ColorX & clr, const Vec3f & v1, const Vec3f & v2, const Vec3f & v3, const Vec3f & v4)
{
	z_draw_state->colors = clr;
	z_draw_state->vertices_3D = 1;
	z_draw_state->dst_quad3d = Quad3D(v1,v2,v3,v4);
	z_InternalDrawMaster(NULL);
}

// 3D image drawing functions:
void GameXClass::DrawImage3D 		  (const ImageX * img, const Vec3f & v, bool worldUnitScale)
{
	z_draw_state->vertices_3D = worldUnitScale?3:2;
	z_draw_state->dst_point3d = v;
	z_InternalDrawMaster(img);
}
void GameXClass::DrawLineImage3D 	  (const ImageX * img, const Vec3f & v1, const Vec3f & v2, bool worldUnitScale)
{
	z_draw_state->src = img;
	z_draw_state->vertices_3D = worldUnitScale ? 5 : 4;
	z_draw_state->dst_quad3d = Quad3D(v1,v2,v1,v2);
	if(z_draw_state->warp_mode == 0) z_draw_state->warp_mode = 2; // auto-perspective-correct
	z_InternalDrawMaster(img);
}
void GameXClass::DrawTexturedPolygon3D (const ImageX * img, const Vec3f & v1, const Vec3f & v2, const Vec3f & v3, const Vec3f & v4)
{
	z_draw_state->src = img;
	z_draw_state->vertices_3D = 1;
	z_draw_state->dst_quad3d = Quad3D(v1,v2,v3,v4);
	if(z_draw_state->warp_mode == 0) z_draw_state->warp_mode = 2; // auto-perspective-correct
	z_InternalDrawMaster(img);
}
void GameXClass::ClearScreen ()
{
	if(z_d3d_main)
	{
		z_EmptyVBuffer();
		z_LockVBuffer();
		z_d3d_device->Clear(0,NULL,(z_d3d_zbuffer_enabled ? D3DCLEAR_ZBUFFER : 0) | D3DCLEAR_TARGET,0,1.0f,0);
		needReClearZ = false;
	}
	else
		z_InternalDrawRectDD(0,0,0, z_win_viewx1,z_win_viewy1,z_win_viewx2,z_win_viewy2);
}
void GameXClass::FillScreen	(const ColorX & clr)
{
	if(z_d3d_main)
	{
		if((((GetDrawMode()|DRAW_PLAIN) & DRAWMODESMASK) != DRAW_PLAIN)
		|| clr.HasTransparency() || memcmp(clr.c[0], clr.c[1],4*sizeof(int)) || memcmp(clr.c[0], clr.c[2],4*sizeof(int)) || memcmp(clr.c[0], clr.c[3],4*sizeof(int)))
		{
			z_draw_state->colors = clr;
			z_draw_state->vertices_3D = 0;
			z_draw_state->dst_quad2d = Quad2D(0,0,DEFAULT,DEFAULT);
			z_InternalDrawMaster(NULL);
		}
		else
		{
			z_EmptyVBuffer();
			z_LockVBuffer();
			z_d3d_device->Clear(0,NULL,(z_d3d_zbuffer_enabled ? D3DCLEAR_ZBUFFER : 0) | D3DCLEAR_TARGET, RGBA_MAKE(clr.c[0][0], clr.c[0][1], clr.c[0][2], 255), 1.0f, 0);
			needReClearZ = false;
		}
	}
	else
	{
		z_InternalDrawRectDD(clr.c[0][0],clr.c[0][1],clr.c[0][2], z_win_viewx1,z_win_viewy1,z_win_viewx2,z_win_viewy2);
	}
}
void GameXClass::BlurScreen	(float amount)
{
	BlurRect(amount, 0,0,DEFAULT,DEFAULT);
}

// fills quadrilateral with screen vertices,
// returns false if it isn't on the screen
bool GameXClass::z_GetScreenVertices(float * x, float * y, float * z)
{
	switch(z_draw_state->vertices_3D)
	{
		case 0: // regular 2D sprite or rectangle or polygon
			{
///				FOR(i,4)
///				{
///					x[i] = z_draw_state->dst_quad2d.x[i];
///					y[i] = z_draw_state->dst_quad2d.y[i];
///					z[i] = z_draw_state->depth2D;
///				}
				z[0] = z[1] = z[2] = z[3] = z_draw_state->depth2D;
				memcpy(x, z_draw_state->dst_quad2d.x, 4*sizeof(float));
				memcpy(y, z_draw_state->dst_quad2d.y, 4*sizeof(float));

				if(x[3] != DEFAULT2 && (z_draw_state->scalex != 1.0f || z_draw_state->scaley != 1.0f))
				{
					float xavg = (x[0]+x[1]+x[2]+x[3])/4.0f;
					float yavg = (y[0]+y[1]+y[2]+y[3])/4.0f;
					FOR(i,4)
					{
						x[i] = ((x[i] - xavg) * z_draw_state->scalex) + xavg;
						y[i] = ((y[i] - yavg) * z_draw_state->scaley) + yavg;
					}
				}
			}
			break;
		case 4: // 3D line (line sprite)
		case 5: // 3D line with scale in world units of width instead of as width factor
			{
				Vec3f top = z_draw_state->dst_quad3d.v[1];
				Vec3f bottom = z_draw_state->dst_quad3d.v[0];
				Vec3f along = top; along -= bottom;
				
				float relative_width;
				if(z_draw_state->vertices_3D == 4)
				{
					relative_width = z_draw_state->scalex/2.0f;
					if(z_draw_state->src != NULL)
					{
						RECT rect = z_ConvertRECT(z_draw_state->src_rect, z_draw_state->src);
						relative_width *= (float)(rect.right-rect.left)/(float)(rect.bottom-rect.top);
					}
				}
				else
				{
					float distance = along.Length();
					if(distance == 0.0f)
						return false;
					relative_width = z_draw_state->scalex / distance;
				}
				if(relative_width == 0.0f) return false;

				relative_width *= along.Length();
				
				Vec3f into = z_draw_state->cam->GetDirVector();
				Vec3f side = Vec3f(along);
				side.CrossWith(into);
				side.Normalize();

				// do special rotation about line axis here
				if(z_draw_state->angle != 0.0f)
				{
					float angle = z_draw_state->angle * -DEGtoRAD;
					float sinAng = sinf(angle);
					float cosAng = cosf(angle);

					// rotate the "side" vector about the arbitrary axis "along":
					along.Normalize();
					side = (side*cosAng) + (along*((1.0f-cosAng)*(side.Dot(along)))) + (along.Cross(side)*sinAng);

					z_draw_state->angle = 0.0f; // angle converted to quad coords; prevent extra 2D rotation
				}

				side *= relative_width;

				z_draw_state->dst_quad3d.v[0] = top+side;
				z_draw_state->dst_quad3d.v[1] = top-side;
				z_draw_state->dst_quad3d.v[2] = bottom+side;
				z_draw_state->dst_quad3d.v[3] = bottom-side;

				z_draw_state->vertices_3D = 1; // line converted to quad
			}
			// no break, must be followed by case 1
		case 1: // quad of 4 Vec3f's   (polygon)
			{
				bool ok = true;
				float xd, yd, zd;
				FOR(i,4)
				{
					z_draw_state->cam->Project(z_draw_state->dst_quad3d.v[i], xd, yd, zd);

					if(zd < z_draw_state->cam->GetNear())
						return false;
//					if(zd > z_draw_state->cam->GetNearClip())
//						ok = true;
//					if(zd < -2000) ok = false;


					x[i] = xd;
					y[i] = yd;
					z[i] = zd;

					if(z_draw_state->warp_mode == 2)
					{
						if(zd != 0.0f)	z_draw_state->warp[i] = 1.0f/(float)zd;
						else			z_draw_state->warp[i] = 1.0f;
					}
				}
				if(!ok) return false;
				z_draw_state->angle = 0.0f;
			}
			break;
		case 2: // 3D point, or
		case 3: // 3D point with 3D scaling    (point sprite)
			{
				float xd, yd, zd;			
				z_draw_state->cam->Project(z_draw_state->dst_point3d, xd, yd, zd);

				if(zd < z_draw_state->cam->GetNearClip())
					return false;

				if(z_draw_state->vertices_3D == 3)
				{
					float div = 2.0f*max(0.02f,zd)/z_draw_state->cam->GetPixelsPerUnit();
					float xp = z_draw_state->scalex/div;
					float yp = z_draw_state->scaley/div;

					x[0] = xd-xp;
					x[1] = xd+xp;
					x[2] = xd-xp;
					x[3] = xd+xp;
					y[0] = yd-yp;
					y[1] = yd-yp;
					y[2] = yd+yp;
					y[3] = yd+yp;
					z[0] = zd;
					z[1] = zd;
					z[2] = zd;
					z[3] = zd;
				} else
				{
					x[0] = xd;
					x[2] = xd;
					x[3] = (float)DEFAULT2;
					y[0] = yd;
					y[1] = yd;
					z[0] = zd;
					z[1] = zd;
					z[2] = zd;
					z[3] = zd;
				}

			}
			break;
		case -1: // 2D line (line sprite)
		case -2: // 2D line with scale in pixels of width instead of as width factor
			{
				float xd[2], yd[2];

				xd[0] = z_draw_state->dst_quad2d.x[0];
				yd[0] = z_draw_state->dst_quad2d.y[0];
				xd[1] = z_draw_state->dst_quad2d.x[1];
				yd[1] = z_draw_state->dst_quad2d.y[2];

				float rise = yd[1]-yd[0];
				float run = xd[0]-xd[1];

				float relative_width;
				if(z_draw_state->vertices_3D == -1)
				{
					relative_width = z_draw_state->scalex/2.0f;
					if(z_draw_state->src != NULL)
					{
						RECT rect = z_ConvertRECT(z_draw_state->src_rect, z_draw_state->src);
						relative_width *= (float)(rect.right-rect.left)/(float)(rect.bottom-rect.top);
					}
				} else
				{
					float distance = sqrtf(run*run+rise*rise);
					if(distance == 0.0f) return false;
					relative_width = z_draw_state->scalex/2.0f / distance;
				}
				if(relative_width == 0.0f) return false;

				rise *= relative_width;
				run *= relative_width;

				x[0] = xd[1] + rise;
				y[0] = yd[1] + run;
				z[0] = z_draw_state->depth2D;

				x[1] = xd[1] - rise;
				y[1] = yd[1] - run;
				z[1] = z_draw_state->depth2D;

				x[2] = xd[0] + rise;
				y[2] = yd[0] + run;
				z[2] = z_draw_state->depth2D;

				x[3] = xd[0] - rise;
				y[3] = yd[0] - run;
				z[3] = z_draw_state->depth2D;
			}
			break;
	}

	if((z_draw_state->scalex <= 0.0f || z_draw_state->scaley <= 0.0f) && !(z_draw_state->flags & DRAWOP_NOCULL))
		return false;

	if(x[3] <= DEFAULT2) // default value that means we have to convert a point into a rectangle
	{
		if(x[3] == DEFAULT2)
		{
			RECT sourceRect = z_ConvertRECT(z_draw_state->src_rect, z_draw_state->src);
			float xOff = z_draw_state->scalex*(sourceRect.right - sourceRect.left);
			float yOff = z_draw_state->scaley*(sourceRect.bottom - sourceRect.top);

			if(z_draw_state->src != NULL && z_draw_state->src != VIEWPORT)
			{
				if(z_draw_state->src->z_xmult != 1.0f)
					xOff /= z_draw_state->src->z_xmult;
				if(z_draw_state->src->z_ymult != 1.0f)
					yOff /= z_draw_state->src->z_ymult;
			}

			x[2] = x[0];
			x[3] = x[0] + xOff;
			x[1] = x[3];
			y[3] = y[0] + yOff;
			y[2] = y[3];

			if(z_draw_state->flags & DRAWOP_CENTERED) // if the point is the center, not the top-left corner, of the destination
			{
				float halfWidth = floorf(((x[1]+x[3]) - (x[0]+x[2])) / 4.0f); // if we don't take the floor, it results in higher accuracy but more blurriness with images that should have 1 to 1 pixel to texel ratio
				float halfHeight = floorf(((y[2]+y[3]) - (y[0]+y[1])) / 4.0f);
				FOR(i,4)
				{
					x[i] -= halfWidth;
					y[i] -= halfHeight;
				}
				z_draw_state->flags &= ~DRAWOP_CENTERED;
			}
		}
		else if(x[3] == DEFAULT) // default value for entire destination
		{
			if(z_draw_state->dst == VIEWPORT)
			{
				x[0] = x[2] = (float)z_win_viewx1;
				y[0] = y[1] = (float)z_win_viewy1;
				x[1] = x[3] = (float)z_win_viewx2;
				y[2] = y[3] = (float)z_win_viewy2;
			}
			else
			{
				x[1] = (float)z_draw_state->dst->z_xres;
				y[2] = (float)z_draw_state->dst->z_yres;
				if(z_draw_state->dst->z_xmult != 1.0f)
					x[1] *= z_draw_state->dst->z_xmult;
				if(z_draw_state->dst->z_ymult != 1.0f)
					y[2] *= z_draw_state->dst->z_ymult;
				x[3] = x[1];
				y[3] = y[2];
			}
		}
	}

	if(z_draw_state->transx || z_draw_state->transy)
	{
		FOR(t,4)
		{
			x[t] += z_draw_state->transx;
			y[t] += z_draw_state->transy;
		}
	}

	float angle = z_draw_state->angle;
	if(angle != 0.0f)
	{
		float x_offset = (x[0]+x[1]+x[2]+x[3]) * 0.25f;
		float y_offset = (y[0]+y[1]+y[2]+y[3]) * 0.25f;
		angle = (360.0f-angle)*DEGtoRAD; // convert to CCW radians
		float cosAngle = cosf(angle);
		float sinAngle = sinf(angle);

		// rotate the image by simply rotating the four points
		// of the rectangular polygon that the image is being mapped onto
		// like so:
		//				   >			  0
		//				  / 			/\				
		//	0----1		 0----1 	   /  \ 			
		//	|	 |		 |	  |\	2 /    \	   
		//	|	 |		 |	  | v	 /		\	   
		//	|	 |	   ^ |	  | 	 \		/		 
		//	|	 |		\|	  | 	  \    / 1		   
		//	2----3	->	 2----3   ->   \  / 				
		//					 /			\/				  
		//					<		   3	 

		FOR(i,4)
		{
			x[i] -= x_offset;
			y[i] -= y_offset;
			float xr = x[i]*cosAngle - y[i]*sinAngle;
			float yr = x[i]*sinAngle + y[i]*cosAngle;
			x[i] = xr + x_offset;
			y[i] = yr + y_offset;
		}
	}

	// catch the things that are completely out of bounds and skip drawing them:
	bool inside = false;
	FOR(i,4)
	{
		if(x[i] >= z_draw_state->dst_bounds.left && x[i] <= z_draw_state->dst_bounds.right && y[i] >= z_draw_state->dst_bounds.top && y[i] <= z_draw_state->dst_bounds.bottom)
		{
			inside = true;
			break;
		}
	}
	if(!inside)
	{
		// none of the points are in bounds, but that doesn't necessarily mean it's safe to skip drawing it...
		// we'll just do a simple check to see if all of the ones that were out of bounds
		// were on the same side, in which case they couldn't possibly intersect the boundary:
		
		int i;
		bool allout;

		allout = true;
		for(i = 0 ; i < 4 ; i++)
		{
			if(x[i] >= z_draw_state->dst_bounds.left)
			{
				allout = false;
				break;
			}
		}
		if(allout) return false; // all points are too far left; don't draw

		allout = true;
		for(i = 0 ; i < 4 ; i++)
		{
			if(y[i] >= z_draw_state->dst_bounds.top)
			{
				allout = false;
				break;
			}
		}
		if(allout) return false; // all points are too far up; don't draw

		allout = true;
		for(i = 0 ; i < 4 ; i++)
		{
			if(x[i] <= z_draw_state->dst_bounds.right)
			{
				allout = false;
				break;
			}
		}
		if(allout) return false; // all points are too far right; don't draw

		allout = true;
		for(i = 0 ; i < 4 ; i++)
		{
			if(y[i] <= z_draw_state->dst_bounds.bottom)
			{
				allout = false;
				break;
			}
		}
		if(allout) return false; // all points are too far down; don't draw
		
		// if we got here, it *probably* needs to be drawn...
		// here is the only no-draw case we may have missed:
		//
		//			polygon (with image on it)
		//			  /\
		//			 /	\
		//			/	 \
		//		   <	  \
		//		   .\	   >
		//		   . \	  /
		// +--------+ \  /
		// |		|..\/
		// |		|
		// |		|
		// |		|
		// +--------+
		// screen (or other destination boundary)
		// 
		// we could do boundary intersection checks at this point but it wouldn't be worth it.
		// since all completely out of bound images would have to go through the check
		// every single time they're drawn, and the time used for intersection checks could add up quickly
	}

	return true;
}

// hack
extern bool z_imageCopyAllowed;

void GameXClass::SetDrawDestination (ImageX* dest)
{
	z_draw_state->dst = dest;

	// shouldn't be changing draw destination or view boundaries during a 3D scene
	assert(z_draw_sort_mode == 0);

	if(dest == VIEWPORT)
	{
		z_draw_state->dst_bounds.left = z_win_viewx1;
		z_draw_state->dst_bounds.top = z_win_viewy1;
		z_draw_state->dst_bounds.right = z_win_viewx2;
		z_draw_state->dst_bounds.bottom = z_win_viewy2;
	}
	else if(dest != NULL)
	{
		if(dest->z_load_queued)
			dest->ForceLoad(); // bring it in
		z_imageCopyAllowed = false;
		dest->ForceLoad(); // invalidate
		z_imageCopyAllowed = true;

		z_draw_state->dst_bounds.left = 0;
		z_draw_state->dst_bounds.top = 0;
		z_draw_state->dst_bounds.right = dest->z_xres;
		z_draw_state->dst_bounds.bottom = dest->z_yres;
	}
}

// convert lowered (from 255) alpha into lowered r,g,b
// used for drawing modes that draw using red, green, blue, effective opacity determined by brightness,
void GameXClass::z_ScaleByAlpha (ColorX & colors)
{
	FOR(i,4)
	{
		if(colors.c[i][3] != 255)
		{
			FOR(j,3)
				colors.c[i][j] = colors.c[i][j] * colors.c[i][3] / 255;
			colors.c[i][3] = 255;
		}
	}
}

void GameXClass::SpecialPauseFade(float blur1, float blur2, int r1, int r2, int g1, int g2, int b1, int b2, int a)
{
z_pushDrawing();
	bool warn = z_win_warning_on_draw; SetWarningOnDraw(false);
	int xv1 = z_win_viewx1, yv1 = z_win_viewy1, xv2 = z_win_viewx2, yv2 = z_win_viewy2;
	SetView(0,0,z_win_width,z_win_height);
	BlurScreen(blur1);
	BlurScreen(blur2);
	ColorX c;
	if(z_win_fullscreen || z_win_screenbpp != 16)
		c = ColorX(GetRandomInt(r1,r2),GetRandomInt(g1,g2),GetRandomInt(b1,b2),a);
	else
		c = ColorX(GetRandomInt(r1,r2),GetRandomInt(g1,g2),GetRandomInt(b1,b2),a,
				   GetRandomInt(r1,r2),GetRandomInt(g1,g2),GetRandomInt(b1,b2),a,
				   GetRandomInt(r1,r2),GetRandomInt(g1,g2),GetRandomInt(b1,b2),a,
				   GetRandomInt(r1,r2),GetRandomInt(g1,g2),GetRandomInt(b1,b2),a);
	FillScreen(c);
	SetView(xv1,yv1,xv2,yv2);
	if(warn) SetWarningOnDraw(true);
z_popDrawing();
}

// blurs the target image or the viewport by scaling down then up
// amount of 0.0f means not blurred at all, 0.5f means somewhat blurred, 1.0f means blurred all the way into a single color
void GameXClass::BlurRect(float amount, int x1, int y1, int x2, int y2)
{
	if(z_win_warning_on_draw && z_win_init_done && z_draw_state->dst == VIEWPORT)
	{
		z_ReportProblem(GAMEX_INVALID_DRAW_TIME_FAIL);
		ResetDrawStates();
		return;
	}

	if(amount <= 0.01f || (!z_win_blur_supported && z_draw_state->dst == VIEWPORT)) return;

	if(x2 <= x1 && x2 != DEFAULT)
	{
		if(z_draw_state->flags & DRAWOP_NOCULL && x1 != x2)
		{
			int t = x2;
			x2 = x1;
			x1 = t;
		} else return;
	}
	if(y2 <= y1 && y2 != DEFAULT)
	{
		if(z_draw_state->flags & DRAWOP_NOCULL && y1 != y2)
		{
			int t = y2;
			y2 = y1;
			y1 = t;
		} else return;
	}
	
	amount = (1.0f - amount) * (1.0f - amount); // invert about 0.5 and adjust amount toward 0
	LPDIRECTDRAWSURFACE7 surface;
	if(z_draw_state->dst == VIEWPORT)
	{
		surface = z_dd_back_buf;
		z_UnlockBackBuffer();
	} else if(z_draw_state->dst == NULL)
	{
		return;
	} else
	{
		surface = z_draw_state->dst->z_surface;
	}

	if(x2 == DEFAULT) { // blur entire thing
		if(z_draw_state->dst == VIEWPORT)
		{
			x1 = z_win_viewx1;
			y1 = z_win_viewy1;
			x2 = z_win_viewx2;
			y2 = z_win_viewy2;
		} else
		{
			x2 = z_draw_state->dst->z_xres;
			y2 = z_draw_state->dst->z_yres;
			if(z_draw_state->dst->z_xmult != 1.0f)
				x2 = (int)((float)x2 * z_draw_state->dst->z_xmult);
			if(z_draw_state->dst->z_ymult != 1.0f)
				y2 = (int)((float)y2 * z_draw_state->dst->z_ymult);
		}
	}

	int width = (x2-x1);
	int height = (y2-y1);
	RECT full_rect = {x1,y1,x2,y2};
	RECT mini_rect = {x1+(int)(width*(1-amount))/2, y1+(int)(height*(1-amount))/2, x1+(int)(width*(1+amount))/2, y1+(int)(height*(1+amount))/2} ;
	if(mini_rect.right == mini_rect.left) mini_rect.left--;
	if(mini_rect.bottom == mini_rect.top) mini_rect.top--;

	z_InternalDrawImageDD(surface, surface, full_rect, mini_rect, 1);
	z_InternalDrawImageDD(surface, surface, mini_rect, full_rect, 1);
}

void GameXClass::DrawText (int x, int y, const char* msg, int r, int g, int b, float fontSize)
{
	PUSH_VBUF();
	if(z_draw_state->flags & DRAWMODESMASK)
	{
		// a mode has been selected - draw with it
		tp.PrintText(msg,(float)x,(float)y,fontSize,false,GameX.GetDrawMode(),ColorX(r,g,b));
	}
	else
	{
		// no mode was selected - draw it embossed since using DRAW_PLAIN as the default would look terrible
		DrawEmbossedText(x, y, msg, r, g, b, 1, 1, fontSize);
	}
	POP_VBUF();
}

void GameXClass::DrawEmbossedText (int x, int y, const char* msg, int r, int g, int b, int xOff, int yOff, float fontSize)
{
	PUSH_VBUF();
	DrawFlags mode = GameX.GetDrawMode() & ~DRAWMODESMASK;
	int val = r*((int)(RED_LUMINANCE*255))/255 + g*((int)(GREEN_LUMINANCE*255))/255 + b*((int)(BLUE_LUMINANCE*255))/255;

	if(val >= 255/4) // add the text to an offset subtractive layer
	{
		tp.PrintText(msg,(float)x+xOff,(float)y+yOff,fontSize,false,mode|DRAW_SUBTRACT,ColorX());

		ColorX c (r,g,b,255, r,g,b,255, r*2/3,g*2/3,b*2/3,255, r*2/3,g*2/3,b*2/3,255);
		tp.PrintText(msg,(float)x,(float)y,fontSize,false,mode|DRAW_ADD,c);
	}
	else // if the color given is dark enough, subtract its inverse from an additive layer instead
	{
		tp.PrintText(msg,(float)x,(float)y,fontSize,false,mode|DRAW_ADD,ColorX(255*2/3,255*2/3,255*2/3));

		ColorX c (r,g,b,255, r,g,b,255, r*2/3,g*2/3,b*2/3,255, r*2/3,g*2/3,b*2/3,255);
		c.Invert();
		c *= 0.5f;
		tp.PrintText(msg,(float)x+xOff,(float)y+yOff,fontSize,false,mode|DRAW_SUBTRACT,c);
	}
	POP_VBUF();
}
void GameXClass::DrawOutlinedText (int x, int y, const char* msg, int r, int g, int b, float fontSize)
{
	PUSH_VBUF();
	DrawFlags mode = GameX.GetDrawMode() & ~DRAWMODESMASK;
	int val = r*((int)(RED_LUMINANCE*255))/255 + g*((int)(GREEN_LUMINANCE*255))/255 + b*((int)(BLUE_LUMINANCE*255))/255;

	static const int dir[8] = {1,1,1,0,-1,-1,-1,0}; // direction table for outline

	if(val >= 255/4) // add the text to an offset subtractive layer
	{
		// Draw a dark outline around the text
		FOR(i,8)
			tp.PrintText(msg,(float)x+dir[i],(float)y+dir[(i+6)%8],fontSize,false,mode|DRAW_SUBTRACT,ColorX());

		// Draw the text
		ColorX c (r,g,b,255, r,g,b,255, r*2/3,g*2/3,b*2/3,255, r*2/3,g*2/3,b*2/3,255);
		tp.PrintText(msg,(float)x,(float)y,fontSize,false,mode|DRAW_ADD,c);
	}
	else // if the color given is dark enough, subtract its inverse from an additive layer instead
	{
		// Draw a bright outline around the text
		FOR(i,8)
			tp.PrintText(msg,(float)x+dir[i],(float)y+dir[(i+6)%8],fontSize,false,mode|DRAW_ADD,ColorX(255*2/3,255*2/3,255*2/3));

		// Draw the text
		ColorX c (r,g,b,255, r,g,b,255, r*2/3,g*2/3,b*2/3,255, r*2/3,g*2/3,b*2/3,255);
		c.Invert();
		c *= 0.5f;
		tp.PrintText(msg,(float)x,(float)y,fontSize,false,mode|DRAW_SUBTRACT,c);
	}
	POP_VBUF();
}

void GameXClass::DrawTextAdvanced (float x, float y, const char * msg, float fontSize, bool centered, int drawMode, ColorX & color, float angle, float yScale, bool rotateBaseline, bool forceFixedWidth)
{
	PUSH_VBUF();
	tp.PrintText(msg,x,y,fontSize,centered,drawMode,color,angle,yScale,rotateBaseline,forceFixedWidth);
	POP_VBUF();
}

void GameXClass::DrawGDIText (int x, int y, const char *msg, int r, int g, int b)
{
	if(z_win_warning_on_draw && z_win_init_done)
	{
		z_ReportProblem(GAMEX_INVALID_DRAW_TIME_FAIL);
		ResetDrawStates();
		return;
	}

	HDC hdc = NULL;

	// in case we don't have DirectDraw
	if(!z_dd_back_buf)
	{
		if(!z_win_hwnd)
			return;

		hdc = GetDC(z_win_hwnd);

		if(!hdc)
			return;
	}
	else
	{
		z_EmptyVBuffer();
		z_LockVBuffer();

		z_LockBackBuffer();

		HRESULT status = z_dd_back_buf->GetDC(&hdc); // Get the device context handle.
		if(FAILED(status))
		{
			z_DirectDrawError(status);
			return;
		}
	}
										// Write the message.
	SetBkMode(hdc, TRANSPARENT);
	SetTextColor(hdc, RGB(r,g,b));
	TextOut(hdc, x, y, msg, (int)(strlen(msg))); // just use the Windows function for now, although it's slow

	// Release the device context.
	if(!z_dd_back_buf)
		ReleaseDC(z_win_hwnd, hdc);
	else
		z_dd_back_buf->ReleaseDC(hdc);		

	return;
}

void GameXClass::DrawOutlinedGDIText (int x, int y, const char *msg, int r, int g, int b, int ro, int go, int bo)
{
	if(z_win_warning_on_draw && z_win_init_done)
	{
		z_ReportProblem(GAMEX_INVALID_DRAW_TIME_FAIL);
		ResetDrawStates();
		return;
	}

	z_EmptyVBuffer();
	z_LockVBuffer();

	z_LockBackBuffer();

	HDC hdc = NULL;
	HRESULT status = z_dd_back_buf->GetDC(&hdc); // Get the device context handle.
	if(FAILED(status))
	{
		z_DirectDrawError(status);
		return;
	}
										// Write the message.
	SetBkMode(hdc, TRANSPARENT);

	// Draw an outline around the text
	const int dir[8] = {1,1,1,0,-1,-1,-1,0}; // direction table for outline
	SetTextColor(hdc, RGB(ro,go,bo));
	FOR(i,8)
		TextOut(hdc, x+dir[i], y+dir[(i+6)%8], msg, (int)(strlen(msg)));

	// Now draw the text normally
	SetTextColor(hdc, RGB(r,g,b));
	TextOut(hdc, x, y, msg, (int)(strlen(msg)));

	z_dd_back_buf->ReleaseDC(hdc);		// Release the device context.
	return;
}


// draws an image tiled over the entire screen,
// good for things such as drawing backgrounds
void GameXClass::DrawTiled (const ImageX * img, float translation_x, float translation_y, float rotation_center_x, float rotation_center_y)
{
	bool clear_states = !(z_draw_state->flags & DRAWOP_KEEPSTATES);
	z_draw_state->flags |= DRAWOP_KEEPSTATES;

	int map_x1, map_y1;
	int map_x2, map_y2; 
	int off_x, off_y;
	float img_x, img_y; 
	float mx, my;

	RECT source_rect = z_ConvertRECT(z_draw_state->src_rect, img);

	float xres = z_draw_state->scalex * (float)(source_rect.right - source_rect.left);
	float yres = z_draw_state->scaley * (float)(source_rect.bottom - source_rect.top);
	if(img != NULL && z_draw_state->src != VIEWPORT)
	{
		if(img->z_xmult != 1.0f)
			xres /= img->z_xmult;
		if(img->z_ymult != 1.0f)
			yres /= img->z_ymult;
	}
	if(xres < 2 || yres < 2) return;

	z_ClipTile ((int)translation_x+z_win_viewx2,(int)translation_y+z_win_viewy2, (int)xres, (int)yres, map_x2, map_y2, off_x, off_y) ;
	z_ClipTile ((int)translation_x+z_win_viewx1,(int)translation_y+z_win_viewy1, (int)xres, (int)yres, map_x1, map_y1, off_x, off_y) ;

	float map_numx = (float)(map_x2 - map_x1 + 1);
	float map_numy = (float)(map_y2 - map_y1 + 1);

///	PUSH_VBUF(); // apparently slower

	if(z_draw_state->angle == 0.0f)
	{
		z_draw_state->vertices_3D = 0;
		for(my = 0; my < map_numy; my++)
		{
			img_y = my * yres - off_y;
			img_x = (float)(-off_x);
			for(mx = 0; mx < map_numx; mx++)
			{
				z_draw_state->dst_quad2d = Quad2D(img_x,img_y,DEFAULT2,DEFAULT2);
				z_InternalDrawMaster(img);
				img_x += xres;
			}
		}
	}
	else
	{
		if(rotation_center_x == DEFAULT)
			rotation_center_x = (float)(z_win_viewx1+z_win_viewx2)/2;
		if(rotation_center_y == DEFAULT)
			rotation_center_y = (float)(z_win_viewy1+z_win_viewy2)/2;

		float sinAngle, cosAngle;

		sinAngle = sinf((360-z_draw_state->angle)*DEGtoRAD);
		cosAngle = cosf((360-z_draw_state->angle)*DEGtoRAD);

		map_numx *= 3;
		map_numy *= 3;

		z_draw_state->vertices_3D = 0;
		for(my = -map_numy; my < map_numy; my++)
		{
			img_y = my * yres;
			img_x = -map_numx * xres;
			for(mx = -map_numx; mx < map_numx; mx++)
			{
				float real_x, real_y, x_offset, y_offset;
				x_offset = rotation_center_x - xres/2;
				y_offset = rotation_center_y - yres/2;
				img_x -= x_offset;
				img_y -= y_offset;
				real_x = img_x * cosAngle - img_y * sinAngle - translation_x;
				real_y = img_x * sinAngle + img_y * cosAngle - translation_y;
				real_x += x_offset;
				real_y += y_offset;
				img_x += x_offset;
				img_y += y_offset;
				z_draw_state->dst_quad2d = Quad2D(real_x,real_y,DEFAULT2,DEFAULT2);
				z_InternalDrawMaster(img);
				img_x += xres;
			}
		}
	}
	if(clear_states)
		ResetDrawStates();

///	POP_VBUF();

	return;
}

//*** Switched Drawing Routines
static void DrawPixel16 (int x, int y, int r, int g, int b, int a = 255); // Sets pixel (16-bit) -- r,g,b,a out of 255
static void DrawPixel32 (int x, int y, int r, int g, int b, int a = 255); // Sets pixel (32-bit) -- r,g,b,a out of 255
static void ReadPixel16 (int x, int y, int &r, int &g, int &b, int &a); // Gets pixel (16-bit) -- r,g,b,a out of 255
static void ReadPixel32 (int x, int y, int &r, int &g, int &b, int &a); // Gets pixel (32-bit) -- r,g,b,a out of 255
// For images that must be scaled due to limitations:
static void DrawScaledPixel16 (int x, int y, int r, int g, int b, int a = 255);
static void DrawScaledPixel32 (int x, int y, int r, int g, int b, int a = 255);
static void ReadScaledPixel16 (int x, int y, int &r, int &g, int &b, int &a);
static void ReadScaledPixel32 (int x, int y, int &r, int &g, int &b, int &a);
// So we can error-check the usage of drawpixel and readpixel:
static void DrawPixelAccessError (int x, int y, int r, int g, int b, int a = 255);
static void DrawPixelConstError (int x, int y, int r, int g, int b, int a = 255);
static void ReadPixelAccessError (int x, int y, int &r, int &g, int &b, int &a);


void GameXClass::AccessPixels (ImageX * destination)
{
	if(destination && destination != VIEWPORT)
	{
		if(destination->z_load_queued)
			destination->ForceLoad(); // bring it in
		z_imageCopyAllowed = false;
		destination->ForceLoad(); // invalidate
		z_imageCopyAllowed = true;
	}

	const ImageX* dest = destination;
	z_AccessPixelsHelper(dest, false);
}

void GameXClass::AccessPixels (const ImageX * destination)
{
	if(destination && destination != VIEWPORT)
		IMG_CONST_ASSERT(destination,"pixels"); 

	z_AccessPixelsHelper(destination, destination && destination != VIEWPORT);
}

void GameXClass::z_AccessPixelsHelper (const ImageX * destination, bool readOnly)
{
	// NOTE: This function stores a non-const pointer to the data in the image,
	//       so I have no idea why the compiler doesn't notice that this violates destination being const.
	//       However, we get readOnly as true if it's const and we use that to avoid violating const-ness,
	//       so as to allow us to propagate proper notification of images being changed beforehand

	assertMsg2(z_win_hwnd, "Trying to access an image (at 0x%08x) before GameX has been initialized.", destination);
	if(!z_win_hwnd) return;
	if(z_win_warning_on_draw && z_win_init_done && (destination == VIEWPORT || destination == NULL))
	{
		z_ReportProblem(GAMEX_INVALID_DRAW_TIME_FAIL);
		ResetDrawStates();
		return;
	}
	if(destination == VIEWPORT || destination == NULL)
	{
		z_win_pixel_dest_surface = z_dd_back_buf;
		z_win_pixel_dest_data = z_win_screen;
		z_win_pixel_dest_pitch = z_win_adjustedpitch;
		z_win_pixel_dest_ytable = z_win_ytable;

		z_EmptyVBuffer();
		z_LockVBuffer();

		z_LockBackBuffer();

		if(z_win_screenbpp == 32)
		{
			z_FuncDrawPixel = &DrawPixel32;
			z_FuncReadPixel = &ReadPixel32;
		}
		else
		{
			z_FuncDrawPixel = &DrawPixel16;
			z_FuncReadPixel = &ReadPixel16;
		}
	}
	else
	{
		z_win_pixel_dest_surface = destination->z_surface;
		
		if(z_win_pixel_dest_surface == NULL)
		{
			z_win_pixel_dest_data = (char *) destination->z_data;
			z_win_pixel_dest_pitch = destination->z_xres*8/z_win_screenbpp;
			z_win_pixel_dest_ytable = destination->z_ytable;

			if(z_win_screenbpp == 32)
			{
				z_FuncDrawPixel = &DrawPixel32;
				z_FuncReadPixel = &ReadPixel32;
			}
			else
			{
				z_FuncDrawPixel = &DrawPixel16;
				z_FuncReadPixel = &ReadPixel16;
			}
		}
		else
		{
			if(destination->z_surf_bpp == 32)
			{
				if(destination->z_xmult == 1.0f && destination->z_ymult == 1.0f)
				{
					z_FuncDrawPixel = &DrawPixel32;
					z_FuncReadPixel = &ReadPixel32;
				}
				else
				{
					z_FuncDrawPixel = &DrawScaledPixel32;
					z_FuncReadPixel = &ReadScaledPixel32;
					z_win_pixel_dest_xmult = destination->z_xmult;
					z_win_pixel_dest_ymult = destination->z_ymult;
				}
			}
			else
			{
				if(destination->z_xmult == 1.0f && destination->z_ymult == 1.0f)
				{
					z_FuncDrawPixel = &DrawPixel16;
					z_FuncReadPixel = &ReadPixel16;
				}
				else
				{
					z_FuncDrawPixel = &DrawScaledPixel16;
					z_FuncReadPixel = &ReadScaledPixel16;
					z_win_pixel_dest_xmult = destination->z_xmult;
					z_win_pixel_dest_ymult = destination->z_ymult;
				}
			}

			DDSURFACEDESC2 ddsd;
			ZeroMemory(&ddsd, sizeof(DDSURFACEDESC2)) ;
			ddsd.dwSize = sizeof(ddsd) ;
			z_DirectDrawError(z_win_pixel_dest_surface->Lock(NULL, &ddsd, DDLOCK_WAIT | DDLOCK_SURFACEMEMORYPTR, NULL));
			z_win_pixel_dest_data = (char *) ddsd.lpSurface;
			z_win_pixel_dest_pitch = ddsd.lPitch * 8 / destination->z_surf_bpp;
			z_win_pixel_dest_ytable = destination->z_ytable;

			if(!readOnly)
				destination->z_non_empty = true;
		}
	}

	if(readOnly)
		z_FuncDrawPixel = &DrawPixelConstError;
}


void GameXClass::z_NullDrawPixelFuncs ()
{
	z_FuncDrawPixel = &DrawPixelAccessError;
	z_FuncReadPixel = &ReadPixelAccessError;
}

void GameXClass::EndPixelAccess ()
{
	if(z_win_pixel_dest_data != z_win_screen && z_win_pixel_dest_surface != NULL)
	{
		z_win_pixel_dest_surface->Unlock(NULL);
	}
	z_NullDrawPixelFuncs();
}

// this function is very slow if compiler optimizations are off (i.e. in debug mode)
// because it is made to be inlined.
static void DrawPixel16 (int x, int y, int r, int g, int b, int a) // r, g, and b are out of 255
{
//	((unsigned short *)GameX.z_win_pixel_dest_data)[x+GameX.z_win_ytable[y]] = (((r >> (8-GameX.z_win_maxr)) << GameX.z_win_shiftr) + ((g >> (8-GameX.z_win_maxg)) << GameX.z_win_shiftg) + ((b >> (8-GameX.z_win_maxb)) << GameX.z_win_shiftb));
	((unsigned short *)GameX.z_win_pixel_dest_data)[x+GameX.z_win_pixel_dest_ytable[y]] = (((r >> 3) << GameX.z_win_shiftr) + ((g >> (8-GameX.z_win_maxg)) << 5) + (b >> 3)); // optimized
}

static void DrawPixel32 (int x, int y, int r, int g, int b, int a) // r, g, and b are out of 255
{
//	((unsigned long *)GameX.z_win_pixel_dest_data)[x+GameX.z_win_ytable[y]] = (((r >> (8-GameX.z_win_maxr)) << GameX.z_win_shiftr) + ((g >> (8-GameX.z_win_maxg)) << GameX.z_win_shiftg) + ((b >> (8-GameX.z_win_maxb)) << GameX.z_win_shiftb));
	((unsigned long *)GameX.z_win_pixel_dest_data)[x+GameX.z_win_pixel_dest_ytable[y]] = ((a << 24) + (r << 16) + (g << 8) + b); // optimized
}

static void ReadPixel16 (int x, int y, int &r, int &g, int &b, int &a)
{
	unsigned short v = ((unsigned short *)GameX.z_win_pixel_dest_data)[x+GameX.z_win_pixel_dest_ytable[y]];
	b =  (v & 0x1F) << 3;
	g = ((v & GameX.z_win_maskg) >> 5) << (8-GameX.z_win_maxg) ;
	r = ((v & GameX.z_win_maskr) >> GameX.z_win_shiftr) << 3;
	a = 255;
}

static void ReadPixel32 (int x, int y, int &r, int &g, int &b, int &a)
{
	unsigned long v = ((unsigned long *)GameX.z_win_pixel_dest_data)[x+GameX.z_win_pixel_dest_ytable[y]];
	b =  v & 0x000000FF;
	g = (v & 0x0000FF00) >> 8;
	r = (v & 0x00FF0000) >> 16;
	a = (v & 0xFF000000) >> 24;
}

static void DrawPixelAccessError (int x, int y, int r, int g, int b, int a)
{
	assertMsg(0, "You must call AccessPixels() before you can call DrawPixel(), and you must call EndPixels() when done. (And you can only access the pixels of one image (or the VIEWPORT) at a time.)");
}

static void DrawPixelConstError (int x, int y, int r, int g, int b, int a)
{
	assertMsg(0, "You cannot call DrawPixel() after calling AccessPixels() on a const ImageX*, because that would change the image.");
}

static void ReadPixelAccessError (int x, int y, int &r, int &g, int &b, int &a)
{
	assertMsg(0, "You must call AccessPixels() before you can call ReadPixel(), and you must call EndPixels() when done. (And you can only access the pixels of one image (or the VIEWPORT) at a time.)");
}

// pixel setter/getter function to be used when the destination is smaller than it's pretending to be

static void DrawScaledPixel16 (int x, int y, int r, int g, int b, int a) // r, g, and b are out of 255
{
	// prevent multiple writes to the same pixel:
	if(GameX.z_win_pixel_dest_xmult != 1.0f && x != 0 && (int)(GameX.z_win_pixel_dest_xmult * (float)(x-1)) == (int)(GameX.z_win_pixel_dest_xmult * (float)(x))) return;
	if(GameX.z_win_pixel_dest_ymult != 1.0f && y != 0 && (int)(GameX.z_win_pixel_dest_ymult * (float)(y-1)) == (int)(GameX.z_win_pixel_dest_ymult * (float)(y))) return;

	// scale the x axis (the y axis is already scaled in the ytable):
	x = Round((float)x * GameX.z_win_pixel_dest_xmult);

	// set the pixel
	((unsigned short *)GameX.z_win_pixel_dest_data)[x+GameX.z_win_pixel_dest_ytable[y]] = (((r >> 3) << GameX.z_win_shiftr) + ((g >> (8-GameX.z_win_maxg)) << 5) + (b >> 3)); // optimized
}

static void DrawScaledPixel32 (int x, int y, int r, int g, int b, int a) // r, g, and b are out of 255
{
	// prevent multiple writes to the same pixel:
	if(GameX.z_win_pixel_dest_xmult != 1.0f && x != 0 && (int)(GameX.z_win_pixel_dest_xmult * (float)(x-1)) == (int)(GameX.z_win_pixel_dest_xmult * (float)(x))) return;
	if(GameX.z_win_pixel_dest_ymult != 1.0f && y != 0 && (int)(GameX.z_win_pixel_dest_ymult * (float)(y-1)) == (int)(GameX.z_win_pixel_dest_ymult * (float)(y))) return;

	// scale the x axis (the y axis is already scaled in the ytable):
	x = Round((float)x * GameX.z_win_pixel_dest_xmult);

	// set the pixel
	((unsigned long *)GameX.z_win_pixel_dest_data)[x+GameX.z_win_pixel_dest_ytable[y]] = ((a << 24) + (r << 16) + (g << 8) + b); // optimized
}

static void ReadScaledPixel16 (int x, int y, int &r, int &g, int &b, int &a)
{
	x = Round((float)x * GameX.z_win_pixel_dest_xmult);
	unsigned short v = ((unsigned short *)GameX.z_win_pixel_dest_data)[x+GameX.z_win_pixel_dest_ytable[y]];
	b =  (v & 0x1F) << 3;
	g = ((v & GameX.z_win_maskg) >> 5) << (8-GameX.z_win_maxg) ;
	r = ((v & GameX.z_win_maskr) >> GameX.z_win_shiftr) << 3;
	a = 255;
}

static void ReadScaledPixel32 (int x, int y, int &r, int &g, int &b, int &a)
{
	x = Round((float)x * GameX.z_win_pixel_dest_xmult);
	unsigned long v = ((unsigned long *)GameX.z_win_pixel_dest_data)[x+GameX.z_win_pixel_dest_ytable[y]];
	b =  v & 0x000000FF;
	g = (v & 0x0000FF00) >> 8;
	r = (v & 0x00FF0000) >> 16;
	a = (v & 0xFF000000) >> 24;
}


void GameXClass::z_InternalDrawMaster (const ImageX * source)
{
	if(z_draw_3D_depth_sort == 2) // replaying drawing
	{
		source = z_draw_state->src;
	}
	else
	{
		if(z_win_err_debug)
		{
			if(z_win_warning_on_draw && z_win_init_done && z_draw_state->dst == VIEWPORT)
			{
				z_ReportProblem(GAMEX_INVALID_DRAW_TIME_FAIL);
				ResetDrawStates();
				return;
			}
		}

		if(!(z_draw_state->flags & DRAWMODESMASK))
			z_draw_state->flags |= DRAW_PLAIN; // assume no drawing mode means plain drawing mode

		if(z_draw_state->dst != VIEWPORT && !(z_draw_state->dst->z_usage & LOAD_TARGETABLE))
			z_draw_state->dst->ConvertUsageTo(z_draw_state->dst->z_usage | LOAD_TARGETABLE); // auto-convert to targetable if needed

		bool immediate;
		if(/*z_draw_3D_depth_sort == 1 &&*/ (z_draw_sort_mode & SCENESORTD))
		{
			// determine whether it needs sorting
			if(!(z_draw_sort_mode & SCENESORTZR) // if not using z-buffer
			|| !(z_draw_state->flags & DRAW_PLAIN) // or if drawing in a transparent mode
			|| z_draw_state->colors.HasTransparency() // or it shading is transparent
			|| (source && source->z_usage & LOAD_ALPHA)) // or if the image contains transparency
			{
				// need to sort
				immediate = false;
			}
			else
			{
				// don't need to sort - can draw immediately
				immediate = true;
			}
		}
		else
		{
			// don't need to sort - can draw immediately
			immediate = true;
		}
		if(z_GetScreenVertices(source,immediate))
		{
			// return if out of bounds or wasn't immediate
			if(!(z_draw_state->flags & DRAWOP_KEEPSTATES))
				ResetDrawStates();
			return;
		}
	}

	RECT sourceRect;

	if(z_draw_state->src != NULL)
	{
		if(source != VIEWPORT && source->z_status != IMG_OK)
		{
			z_ReportProblem(GAMEX_UNINIT_IMG_DRAW_FAIL);
			return;
		}

		sourceRect = z_ConvertRECT(z_draw_state->src_rect, z_draw_state->src);
	}


	z_draw_state->colors.CapComponents();

	if(!(z_draw_state->flags & DRAWOP_NOBLENDALPHA))
	{
		if(!(z_draw_state->flags & DRAW_PLAIN))
		{
			// scale the luminosity by the alpha which would otherwise be ignored
			if(z_draw_state->dst == VIEWPORT || !(z_draw_state->dst->z_usage & LOAD_ALPHA))
			{
				z_ScaleByAlpha(z_draw_state->colors);

				if(!(z_draw_state->colors.HasColor())
				&& z_draw_state->color_effect == D3DTOP_MODULATE)
				{
					if(!(z_draw_state->flags & DRAWOP_KEEPSTATES))
						ResetDrawStates();
					return; // return if there isn't any luminosity
				}
			}
		}
		else
		{
			if(!(z_draw_state->colors.HasOpacity()))
			{
				if(!(z_draw_state->flags & DRAWOP_KEEPSTATES))
					ResetDrawStates();
				return; // return if there isn't any opacity
			}
		}
	}

	int sourceHasSurface, destHasSurface;
	bool sourceHasData, sourceHasAlpha, destHasData, destHasAlpha, destHasDevice;
	LPDIRECTDRAWSURFACE7 p_src_surface, p_dst_surface;
	LPDIRECT3DDEVICE7 p_dst_device;
	unsigned char * p_src_data;
	unsigned char * p_src_alpha;
	unsigned char * p_dst_data;
	unsigned char * p_dst_alpha;
	RECT src_bounds = {0,0,0,0};
	int src_pitch, dst_pitch;

	if(z_draw_state->src == VIEWPORT)
	{
		sourceHasSurface = 1;
		p_src_surface = z_dd_back_buf;
		sourceHasAlpha = false;
		src_pitch = z_win_adjustedpitch;
		src_bounds.right = z_win_width;
		src_bounds.bottom = z_win_height;
	} else if(z_draw_state->src != NULL)
	{
		if(z_draw_state->src->z_status != IMG_OK) return;
		if(z_draw_state->src->z_hassurface)
		{
			sourceHasSurface = z_draw_state->src->z_hassurface;
			p_src_surface = z_draw_state->src->z_surface;
		}
		else sourceHasSurface = 0;

		if(z_draw_state->src->z_data != NULL)
		{
			sourceHasData = true;
			p_src_data = z_draw_state->src->z_data;
		}
		else
			sourceHasData = false;

		if(z_draw_state->src->z_alpha != NULL)
		{
			sourceHasAlpha = true;
			p_src_alpha = z_draw_state->src->z_alpha;
		}
		else
			sourceHasAlpha = false;

		src_pitch = z_draw_state->src->z_xres; // may have to change this line to do depths other than 16 bit
		src_bounds.right = z_draw_state->src->z_xres;
		src_bounds.bottom = z_draw_state->src->z_yres;
		if(z_draw_state->src->z_xmult != 1.0f) src_bounds.right = (LONG)((float)src_bounds.right * z_draw_state->src->z_xmult);
		if(z_draw_state->src->z_ymult != 1.0f) src_bounds.bottom = (LONG)((float)src_bounds.bottom * z_draw_state->src->z_ymult);
	}
	else
	{
		sourceHasSurface = 2;
		p_src_surface = NULL;
		sourceHasData = false;
		sourceHasAlpha = false;
	}

	if(z_draw_state->dst == VIEWPORT)
	{
		destHasSurface = 3;
		p_dst_surface = z_dd_back_buf;

		destHasDevice = true;
		p_dst_device = z_d3d_device;

		destHasAlpha = false;

		dst_pitch = z_win_adjustedpitch;
	}
	else
	{
		if(z_draw_state->dst->z_status != IMG_OK) return;
		if(z_draw_state->dst->z_hassurface)
		{
			destHasSurface = z_draw_state->dst->z_hassurface;
			p_dst_surface = z_draw_state->dst->z_surface;
		}
		else
		{
			dst_pitch = z_draw_state->dst->z_xres; // may have to change this line to do depths other than 16 bit...maybe
			destHasSurface = 0;
		}

		if(z_draw_state->dst->z_data != NULL)
		{
			destHasData = true;
			p_dst_data = z_draw_state->dst->z_data;
		}
		else
			destHasData = false;

		if(z_draw_state->dst->z_alpha != NULL)
		{
			destHasAlpha = true;
			p_dst_alpha = z_draw_state->dst->z_alpha;
		}
		else
			destHasAlpha = false;

		if(z_draw_state->dst->z_device != NULL)
		{
			destHasDevice = true;
			p_dst_device = z_draw_state->dst->z_device;
		}
		else
			destHasDevice = false;
	}

	// NOTE: support for all advanced features is not complete for if we can't draw using 3D,
	//		 but that *should* only happen when drawing FROM the back buffer to somewhere,

	int ok = 0;
	if(sourceHasSurface == 2 && destHasDevice)
	{
		if(p_src_surface)
			ok = z_InternalDrawImageD3D(p_dst_device, p_src_surface, sourceRect, src_bounds, z_draw_state->x, z_draw_state->y, z_draw_state->z, z_draw_state->src->z_surf_width, z_draw_state->src->z_surf_height);
		else
			ok = z_InternalDrawImageD3D(p_dst_device, NULL, FULLRECT, FULLRECT, z_draw_state->x, z_draw_state->y, z_draw_state->z, z_win_width, z_win_height);
	}
	if(!ok) // D3D FAILED or won't work
	{
		float * x, * y, * z;
		x = z_draw_state->x; // pointer to an array of 4 variable-purpose x screen coordinates
		y = z_draw_state->y; // pointer to an array of 4 variable-purpose y screen coordinates
		z = z_draw_state->z; // used for depth sorting

///		bool can_2D_accelerate = (sourceHasSurface && destHasSurface && p_src_surface != NULL && (z_draw_state->flags & DRAW_PLAIN) && alphatype != 2 && z_draw_state->angle == 0.0f && !z_draw_state->colors.HasShading() && (z_draw_state->vertices_3D == 0||z_draw_state->vertices_3D == 2||z_draw_state->vertices_3D == 3));
		bool can_2D_accelerate = destHasSurface ? true : false;

		RECT destRect;
		destRect.left = (LONG)(x[0]+x[2])/2;
		destRect.top = (LONG)(y[0]+y[1])/2;
		destRect.right = (LONG)(x[1]+x[3])/2;
		destRect.bottom = (LONG)(y[2]+y[3])/2;

		if(can_2D_accelerate)
		{
			if(z_draw_state->src == VIEWPORT || z_draw_state->dst == VIEWPORT)
				z_UnlockBackBuffer();
			if(z_dd_was_lost)
			{
				if(z_draw_state->src && z_draw_state->src != VIEWPORT) {p_src_surface = z_draw_state->src->z_surface;}
				if(z_draw_state->dst && z_draw_state->dst != VIEWPORT) {p_dst_surface = z_draw_state->dst->z_surface;}
			}
			if(sourceHasSurface && p_src_surface != NULL) // plain directdraw draw
			{
				int alphatype = 0;
				if(z_draw_state->src != VIEWPORT && z_draw_state->src != NULL)
				{
					if((z_draw_state->src->z_usage & LOAD_MASKED))
						alphatype = 1;
					else if((z_draw_state->src->z_usage & LOAD_ALPHA))
					{
						if(z_draw_state->dst == VIEWPORT || !(z_draw_state->dst->z_usage & LOAD_ALPHA))
						alphatype = 2; // alpha channel in image
					}
				}

				ok = z_InternalDrawImageDD(p_src_surface, p_dst_surface, sourceRect, destRect, !alphatype);
			}
			else // solid fill color
			{
				z_InternalDrawRectDD(z_draw_state->colors.c[0][0],z_draw_state->colors.c[0][1],z_draw_state->colors.c[0][2],
									destRect.left, destRect.top, destRect.right, destRect.bottom);
				ok = true;
			}
		}
		if(!ok)
		{
			// insert software rendering backup code here
		}
	}

	if(z_draw_state->dst != VIEWPORT && destHasSurface == 2)
	{
		// if the destination is a texture, we have to tell D3D to refresh it by
		// locking and unlocking it, or else the image won't change after the first time it's drawn.
		// note that this makes drawing to somewhere else than the screen even slower...
		DDSURFACEDESC2 ddsd;
		ddsd.dwSize = sizeof(ddsd);
		p_dst_surface->Lock(NULL, &ddsd, DDLOCK_WAIT, NULL);
		p_dst_surface->Unlock(NULL);
	}
	if(z_draw_3D_depth_sort != 2)
		if(!(z_draw_state->flags & DRAWOP_KEEPSTATES))
			ResetDrawStates();

	if(z_win_request_slow_debug_drawing && z_win_init_done && z_draw_state->dst == VIEWPORT)
	{
		if((GetAsyncKeyState(VK_ESCAPE) & 0x8000))
		{
			z_win_request_slow_debug_drawing = false; // must be false before choice dialog
			if(ChoiceDialog("Stop drawing slowly?") != 1)
				z_win_request_slow_debug_drawing = true;
		}
		else
		{
			UpdateView();
			Sleep(60);
		}
	}
} // end of z_InternalDrawMaster

static D3DVIEWPORT7 view;
int lastSrcBlend = -1, lastDestBlend = -1;
int lastShadeCombineEffect = -1, lastColorShift = -1;
int lastColorKeyEnable = -1, lastInvertEnable = -1, lastCullEnable = -1, lastFilterEnable = -1, lastDitherEnable = -1, lastAlphaBlendEnable = -1;
bool work_burn, work_inten;
int srcBlend, destBlend, color_shift;
D3DTEXTUREOP shade_combine_effect;
int colorKeyEnable, invertEnable, cullEnable, filterEnable, ditherEnable, alphaBlendEnable;
int alphatype = 0;

void GameXClass::z_ApplyDrawModeWorkarounds (LPDIRECTDRAWSURFACE7 image, bool & work_burn, bool & work_inten)
{
	// Determine which workarounds, if any, we need to apply right now:
	     work_inten = z_win_intensify_workaround	 == 1 && z_draw_state->flags & DRAW_INTENSIFY;
	bool work_subt1 = z_win_subtract_workaround	 == 1 && z_draw_state->flags & DRAW_SUBTRACT && alphatype != 2;
	bool work_subt2 = z_win_subtract_workaround	 == 2 && z_draw_state->flags & DRAW_SUBTRACT;
	bool work_alpha = z_win_alpha_blend_workaround == 1 && z_draw_state->flags & DRAW_PLAIN && alphatype != 2 && z_draw_state->colors.HasTransparency();
	bool work_ghost = z_win_ghost_workaround		 == 1 && z_draw_state->flags & DRAW_GHOST;
	     work_burn	= z_win_burn_workaround		 == 1 && z_draw_state->flags & DRAW_BURN;
	bool work_addso = z_win_addsoft_workaround	 == 1 && z_draw_state->flags & DRAW_ADDSOFT;
	bool work_addsh = z_win_addsharp_workaround	 == 1 && z_draw_state->flags & DRAW_ADDSHARP;

	// Implement work-around methods as necessary:
	// (these are to work around problems with older video cards, as decided in z_DoInitTests())
	if(work_inten)
	{
		z_draw_state->flags ^= DRAW_ADD | DRAW_INTENSIFY; // switch from intensify to regular additive mode
		z_draw_state->colors /= 2.0f; // cut brightness in half
	}
	if(work_addso)
		z_draw_state->flags ^= DRAW_ADD | DRAW_ADDSOFT; // switch from soft add to regular additive mode
	if(work_addsh)
		z_draw_state->flags ^= DRAW_ADD | DRAW_ADDSHARP; // switch from sharp add to regular additive mode
	if(work_burn) // note that part of the burn workaround happens later, also
		z_draw_state->flags ^= DRAW_INVERT | DRAW_BURN; // switch from burn to inversion
	if(work_subt1 || work_subt2 || work_ghost || work_alpha)
	{
		if(image == NULL || image == z_dd_back_buf)
		{
			if(work_subt1 || work_subt2)
			{
				image = NULL;
				z_draw_state->flags ^= DRAW_PLAIN | DRAW_SUBTRACT; // switch from subtractive to regular alpha mode
				FOR(i,4)
				{
					z_draw_state->colors.c[i][3] = (int)((float)z_draw_state->colors.c[i][0]*RED_LUMINANCE + (float)z_draw_state->colors.c[i][1]*GREEN_LUMINANCE + (float)z_draw_state->colors.c[i][2]*BLUE_LUMINANCE) * z_draw_state->colors.c[i][3]/255;
					z_draw_state->colors.c[i][2] = 0;
					z_draw_state->colors.c[i][1] = 0;
					z_draw_state->colors.c[i][0] = 0;
				}
			}
		} else
		{
			int compat_index = (work_ghost) ? 2 : ((work_subt2) ? 1 : 0);
			ImageX * work_img = z_draw_state->src->z_internal_compatibility_img[compat_index];
			if(work_img == NULL)
			{
				z_SetStatus(GAMEX_CREATING_TEST_SURFACE); // prevent from registering as user-created
				work_img = new ImageX(); // FIXME: when should this get deleted?
				z_draw_state->src->z_internal_compatibility_img[compat_index] = work_img;
				work_img->Create(z_draw_state->src->z_xres, z_draw_state->src->z_yres, true);
				z_SetStatus(GAMEX_READY);
				z_draw_state->src->CopyTo(work_img);
				if(work_subt2 || work_ghost)
				{
					AccessPixels(work_img);
					for(int y = 0 ; y < work_img->z_yres ; y++)
					{
						for(int x = 0 ; x < work_img->z_xres ; x++)
						{
							int r,g,b,a;
							ReadPixel(x,y,r,g,b,a);
							a = (int)((float)r*RED_LUMINANCE + (float)g*GREEN_LUMINANCE + (float)b*BLUE_LUMINANCE) * a/255;
							if(work_subt2) r = 0, g = 0, b = 0;
							DrawPixel(x,y,r,g,b,a);
						}
					}
					EndPixelAccess();
				}
			}
			image = work_img->z_surface;
			if(work_subt2 || work_ghost)
			{
				z_draw_state->flags |= DRAWMODESMASK, z_draw_state->flags ^= DRAWMODESMASK; // turn off all draw modes
				alphatype = 2, z_draw_state->flags | DRAW_PLAIN; // turn on alpha with plain drawing mode
				FOR(i,4) // scale shading transparency with brightness
					z_draw_state->colors.c[i][3] = min(255, (int)((float)z_draw_state->colors.c[i][0]*RED_LUMINANCE + (float)z_draw_state->colors.c[i][1]*GREEN_LUMINANCE + (float)z_draw_state->colors.c[i][2]*BLUE_LUMINANCE) *2* z_draw_state->colors.c[i][3]/255);
			}
		}
	}
}


void GameXClass::z_ProcessDrawMode(int & srcBlend, int & destBlend, int & color_shift)
{
	switch(z_draw_state->flags & DRAWMODESMASK)
	{
		default:
		case DRAW_PLAIN: // plain, masked, or alpha drawing
			if(alphatype != 2 && !z_draw_state->colors.HasTransparency())
			{
				srcBlend = D3DBLEND_ONE; // Cs = 1
				destBlend = D3DBLEND_ZERO; // Cd = 0
				// so C = source * Cs + dest * Cd
				//		= source * (1) + dest * (0)
				//		= source
			}
			else
			{
				srcBlend = D3DBLEND_SRCALPHA; // Cs = alpha
				destBlend = D3DBLEND_INVSRCALPHA; // Cd = 1-alpha
				// so C = source * Cs + dest * Cd
				//		= source * (alpha) + dest * (1-alpha)
				//		= dest + source*alpha - dest*alpha
			}
			return;
		case DRAW_ADD: // additive -- r, g, b at 255 is full additive, reduce r for cyan-tinted added, etc.
			if(alphatype != 2)
				srcBlend = D3DBLEND_ONE; // Cs = 1, or
			else
				srcBlend = D3DBLEND_SRCALPHA; // Cs = alpha

			destBlend = D3DBLEND_ONE; // Cd = 1
			// so C = source * Cs + dest * Cd
			//		= source * (1 [* alpha]) + dest * (1)
			//		= dest + source [* alpha]
			return;
		case DRAW_SUBTRACT: // subtractive -- r, g, b at 255 is full subtractive, reduce r to subtract less red, etc.
			// note: not totally accurate subtraction because this does not use DirectX 8.0
			//		 so the closest thing in DX7 is this, the inverse of DRAW_INTENSIFY
			srcBlend = D3DBLEND_ZERO; // Cs = 0, or
			destBlend = D3DBLEND_INVSRCCOLOR; // Cd = 1-source
			// so C = source * Cs + dest * Cd
			//		= source * (0) + dest * (1-source)
			//		= dest - source*dest
			color_shift++; // increase effect 2x because it's usually too pale otherwise (so we don't have to render it twice)
			//	  C = dest - (2*source)*dest
			return;
		case DRAW_GHOST: // color varying alpha blend -- r, g, b determine amount, produces interesting auto-alpha effect
			srcBlend = D3DBLEND_SRCCOLOR; // Cs = source
			destBlend = D3DBLEND_INVSRCCOLOR; // Cd = 1-source
			// so C = source * Cs + dest * Cd
			//		= source * (source) + dest * (1-source)
			//		= dest + source*source - source*dest
			return;
		case DRAW_INVERT: // invert by the amount of color in the source -- r,g,b at 255 is full inversion, otherwise hard to describe
			srcBlend = D3DBLEND_INVDESTCOLOR; // Cs = 1-dest
			destBlend = D3DBLEND_INVSRCCOLOR; // Cd = 1-source
			// so C = source * Cs + dest * Cd
			//		= source * (1-dest) + dest * (1-source)
			//		= dest + 2*source - 2*source*dest
			return;
		case DRAW_MULTIPLY: // multiply source by destination
			srcBlend = D3DBLEND_ZERO; // Cs = 0
			destBlend = D3DBLEND_SRCCOLOR; // Cd = source
			// so C = source * Cs + dest * Cd
			//		= source * (0) + dest * (source)
			//		= dest * source
			return;
		case DRAW_BURN: // similar to invert but peaks at midway brightness
			srcBlend = D3DBLEND_INVSRCCOLOR; // Cs = 1-source
			destBlend = D3DBLEND_INVSRCCOLOR; // Cd = 1-source
			// so C = source * Cs + dest * Cd
			//		= source * (1-source) + dest * (1-source)
			//		= source - source*source + dest - source*dest
			return;
		case DRAW_INTENSIFY: // add, but darker as destination becomes darker
			srcBlend = D3DBLEND_DESTCOLOR; // Cs = dest
			destBlend = D3DBLEND_ONE; // Cd = 1
			// so C = source * Cs + dest * Cd
			//		= source * (dest) + dest * (1)
			//		= dest + source*dest
			color_shift++; // increase effect 2x because it's usually too pale otherwise (so we don't have to render it twice)
			//	  C = dest + (2*source)*dest
			return;
		case DRAW_ADDSHARP: // similar to DRAW_ADD but colors show more sharply (increases contrast)
			srcBlend = D3DBLEND_SRCCOLOR; // Cs = source
			destBlend = D3DBLEND_ONE; // Cd = 1
			// so C = source * Cs + dest * Cd
			//		= source * (source) + dest * (1)
			//		= dest + source*source
			color_shift++; // increase effect 2x to compensate for brightness lost in contrast change
			//	  C = dest + (2*source)*(2*source)
			return;
		case DRAW_ADDSOFT: // similar to DRAW_ADD but adds less to brighter areas (prevents overexposure)
			srcBlend = D3DBLEND_INVDESTCOLOR; // Cs = 1-dest
			destBlend = D3DBLEND_ONE; // Cd = 1
			// so C = source * Cs + dest * Cd
			//		= source * (1-dest) + dest * (1)
			//		= dest + source*(1-dest)
			//		= dest + source - source*dest
			return;
	}
}

// Draws a 2D image using 3D acceleration and features
int GameXClass::z_InternalDrawImageD3D (LPDIRECT3DDEVICE7 device, LPDIRECTDRAWSURFACE7 image, RECT & src_rect, RECT & src_bounds, float * x, float * y, float * z, int surf_width, int surf_height)
{
	if(device == NULL) return 0;

	if(z_draw_state->dst == VIEWPORT)
		z_UnlockBackBuffer(); // must be unlocked to draw with D3D (or DD for that matter)

	// debug overdraw testing
	if(debug.IsRunningDebug() && GameX.IsKeyDown(KEY_NUMLOCK))
	{
		image = NULL;
		z_draw_state->src = NULL;
		z_draw_state->colors = ColorX(8,12,16);
		z_draw_state->flags &= ~DRAWMODESMASK;
		z_draw_state->flags |= DRAW_ADD | DRAWOP_NODITHER;
	}


	if(z_d3d_device != NULL
		&& (z_draw_state->dst_bounds.left != view.dwX // it turns out that SetViewport() is really quite slow,
		 || z_draw_state->dst_bounds.top != view.dwY  // so these extra checks can greatly improve performance
		 || z_draw_state->dst_bounds.right - z_draw_state->dst_bounds.left != view.dwWidth // when optimizations are on
		 || z_draw_state->dst_bounds.bottom - z_draw_state->dst_bounds.top != view.dwHeight)) // especially when drawing lots of things
	{
		view.dwX = z_draw_state->dst_bounds.left;
		view.dwY = z_draw_state->dst_bounds.top;
		view.dwWidth = z_draw_state->dst_bounds.right - z_draw_state->dst_bounds.left;
		view.dwHeight = z_draw_state->dst_bounds.bottom - z_draw_state->dst_bounds.top; 
		view.dvMinZ = 0.0f;
		view.dvMaxZ = 1.0f;
		z_d3d_device->SetViewport(&view);
	}
	// Because Direct3D does not support quadrilateral primitives,
	// create vertices for our quad with points ordered like so:
	//
	//	0----1
	//	.	/.
	//	.  / .
	//	. /  .
	//	./	 .
	//	2----3
	//
	// to allow for D3D to use it for a triangle strip.
/*
	float angle = z_draw_state->angle;
	int angle_adjust_counter = 999; // prevent infinite angle adjusting when divide-by-zero is passed as angle
	while(angle >= 360.0f && angle_adjust_counter) angle -= 360.0f, angle_adjust_counter--; // restrict to 0-360 range to allow easier angle checking
	while(angle < 0.0f	  && angle_adjust_counter) angle += 360.0f, angle_adjust_counter--;
	int rotate;
	if(angle==	0.0f || angle== 45.0f || angle== 90.0f || angle == 135.0f
	|| angle == 180.0f || angle == 225.0f || angle == 270.0f || angle == 315.0f)
		rotate = 0; // if the angle is exactly a multiple of 45, we don't have to adjust its inset as much
	else
		rotate = 1;
*/
	float leftMap, topMap, rightMap, bottomMap;

	if(&src_bounds != &FULLRECT)
	{
/*
		if(src_rect.right - src_rect.left == src_bounds.right - src_bounds.left)
		{
			x_adjust = 0.0005f;
		} else
		{
			float scalex = (float)(x[3]-x[0])/(src_rect.right-src_rect.left);
			x_adjust = ((scalex <= 1.0f) ? (0.0005f) : (0.525f-0.5f/scalex));
			if(rotate) x_adjust = min(0.5f, x_adjust+0.25f);
		}
		if(src_rect.bottom - src_rect.top == src_bounds.bottom - src_bounds.top)
		{
			y_adjust = 0.0005f;
		} else
		{
			float scaley = (float)(y[3]-y[0])/(src_rect.bottom-src_rect.top);
			y_adjust = ((scaley <= 1.0f) ? (0.0005f) : (0.525f-0.5f/scaley));
			if(rotate) y_adjust = min(0.5f, y_adjust+0.25f);
		}
*/
		// respond to mirroring request flags
		if(z_draw_state->flags & (DRAWOP_VMIRROR | DRAWOP_HMIRROR))
		{ 
			float x_adjust = 0.5f, y_adjust = 0.5f; // source inset amount
			if(z_draw_state->flags & DRAWOP_VMIRROR)
			{
				// flip the top and bottom mapping values if we want to mirror the image vertically
				const int temp = src_rect.top;
				src_rect.top = src_rect.bottom;
				src_rect.bottom = temp;
				y_adjust = -y_adjust;
			}
			else
			{
				// flip the left and right mapping values if we want to mirror the image horizontally
				const int temp = src_rect.left;
				src_rect.left = src_rect.right;
				src_rect.right = temp;
				x_adjust = -x_adjust;
			}
			leftMap   = ((float)src_rect.left   + x_adjust) / surf_width;	// determine source mapping values
			topMap	  = ((float)src_rect.top    + y_adjust) / surf_height; // (the pixel adjustments fix a
			rightMap  = ((float)src_rect.right  - x_adjust) / surf_width;	//	"sometimes off by one" bug)
			bottomMap = ((float)src_rect.bottom - y_adjust) / surf_height;
		}
		else
		{
			leftMap   = ((float)src_rect.left   + 0.5f) / surf_width;	
			topMap	  = ((float)src_rect.top    + 0.5f) / surf_height;
			rightMap  = ((float)src_rect.right  - 0.5f) / surf_width;	
			bottomMap = ((float)src_rect.bottom - 0.5f) / surf_height;
		}

	}
	else
	{
		leftMap  = 0.0f; topMap    = 0.0f;
		rightMap = 1.0f; bottomMap = 1.0f;
	}

	// determine what the render state needs to be
	z_DetermineStates(image);

	// handle possible render target change and/or first draw change
	bool new_batch = false;
	{
		HRESULT status = D3D_OK;
		if(z_draw_render_batch_device != device)
		{
			if(z_draw_render_batch_open == true)
				status = z_EndRenderBatch(); // sets z_draw_render_batch_open = false
			z_draw_render_batch_device = device;
		}
		if(z_draw_render_batch_open == false)
		{
			if(SUCCEEDED(status))
			{
				status = device->BeginScene();
				if(SUCCEEDED(status))
				{
					z_draw_render_batch_open = true;
					new_batch = true;
				}
			}
		}
	}

	bool somethingChanged = new_batch
		|| image != z_last_texture_surface
		|| lastSrcBlend != srcBlend
		|| lastDestBlend != destBlend 
		|| lastCullEnable != cullEnable 
		|| lastInvertEnable != invertEnable 
		|| lastFilterEnable != filterEnable 
		|| lastDitherEnable != ditherEnable 
		|| lastColorKeyEnable != colorKeyEnable 
		|| lastAlphaBlendEnable != alphaBlendEnable 
		|| lastShadeCombineEffect != shade_combine_effect
		|| lastColorShift != color_shift;

	// first flush previous draws before changing the state
	if((somethingChanged || z_d3d_vbufferfilled+6 >= z_d3d_vbuffermaxverts)
	&& (z_d3d_vbuffer != NULL && z_d3d_vbuffermem != NULL && z_draw_state->dst == VIEWPORT && z_win_status == GAMEX_READY))
	{
		z_EmptyVBuffer();
		z_LockVBuffer();
	}

	// apply changes to the render state
	if(somethingChanged)
		z_ApplyStates(image, device, new_batch);


	// now for the important part

	static D3DTLVERTEX quad[5];
	{
		if(z_d3d_vbufferfilled != 0)
		{
			quad[4].sx = (*(z_d3d_vbuffermem-1)).sx;
			quad[4].sy = (*(z_d3d_vbuffermem-1)).sy;
		}

		const int signedZ = z_draw_sort_mode & SCENESORTS;
		const float & near_plane = z_draw_state->cam->GetNear();
		const float & z_plane_sep = z_draw_state->cam->GetInvPlaneSep();
		FOR(i,4)
		{
			quad[i].sx = x[i] - 0.5f;
			quad[i].sy = y[i] - 0.5f;

			if(signedZ)
				if(z_win_alternator)
					quad[i].sz = 1.0f - (z[i]-near_plane)*z_plane_sep*0.5f; // for signed z-buffer
				else
					quad[i].sz = (z[i]-near_plane)*z_plane_sep*0.5f; // for signed z-buffer
			else
				quad[i].sz = (z[i]-near_plane)*z_plane_sep; // for normal z-buffer

			quad[i].rhw = (z_draw_state->warp_mode != 0) ? z_draw_state->warp[i] : 1.0f;
			quad[i].color = RGBA_MAKE(z_draw_state->colors.c[i][0], z_draw_state->colors.c[i][1], z_draw_state->colors.c[i][2], z_draw_state->colors.c[i][3]);
			quad[i].tu = (i == 0 || i == 2) ? (leftMap) : (rightMap);
			quad[i].tv = (i == 0 || i == 1) ? (topMap) : (bottomMap);
		}


		if(z_d3d_vbufferenabled && z_d3d_vbuffer != NULL && z_d3d_vbuffermem != NULL && z_draw_state->dst == VIEWPORT && z_win_status == GAMEX_READY)
		{
			// optimization
			// renders into a single continuing triangle strip (using the vertex buffer)

			if(z_d3d_vbufferfilled == 0)
			{
				FOR(i,4)
				{
					FVFTLVERTEX & vertex = *z_d3d_vbuffermem;

					// apparently this way of copying is faster than anything else when optimizations are on
					vertex.sx = quad[i].sx;
					vertex.sy = quad[i].sy;
					vertex.sz = quad[i].sz;
					vertex.rhw = quad[i].rhw;
					vertex.color = quad[i].color;
					vertex.tu = quad[i].tu;
					vertex.tv = quad[i].tv;

	//				z_d3d_vbuffermem += sizeof(FVFTLVERTEX);
					z_d3d_vbuffermem ++;
				}
				z_d3d_vbufferfilled += 4;
			}
			else
			{
				const static int triangleListIndices [6] = {4,0,0,1,2,3}; // triangle strip segment with degenerates on each end

				FOR(j,6)
				{
					int i = triangleListIndices[j];

					FVFTLVERTEX & vertex = *z_d3d_vbuffermem;

					// apparently this way of copying is faster than anything else when optimizations are on
					vertex.sx = quad[i].sx;
					vertex.sy = quad[i].sy;
					if(j >= 2)
					{
						vertex.sz = quad[i].sz;
						vertex.rhw = quad[i].rhw;
						vertex.color = quad[i].color;
						vertex.tu = quad[i].tu;
						vertex.tv = quad[i].tv;
					}

					z_d3d_vbuffermem++;
				}
				z_d3d_vbufferfilled += 6;
			}
		}
		else
		{
			// render immediately if we can't use the vertex buffer

			device->DrawPrimitive(D3DPT_TRIANGLESTRIP, D3DFVF_TLVERTEX, quad,4,0);
			totalAlone++;
		}

	}

	// HACK:
	// finish some workarounds here
	if(z_win_any_draw_mode_workaround)
	{
		// rest of DRAW_BURN workaround is implemented here
		if(work_burn)
		{
			if(z_win_subtract_workaround != 2 && (z_win_subtract_workaround != 1 || alphatype == 2))
			{
				srcBlend = D3DBLEND_ZERO; // Cs = 0, or
				destBlend = D3DBLEND_INVSRCCOLOR; // Cd = 1-source
				device->DrawPrimitive(D3DPT_TRIANGLESTRIP, D3DFVF_TLVERTEX, quad,4,0); // draw again with DRAW_SUBTRACT:
				totalAlone++;
			}
			else // alternate implementation is necessary if the DRAW_SUBTRACT workaround is also needed:
			{
				z_draw_state->flags ^= DRAW_SUBTRACT | DRAW_INVERT; // switch from inversion to subtraction
				z_draw_state->colors /= 2.0f;
				z_InternalDrawImageD3D(device, image, src_rect, src_bounds, x, y, z, surf_width, surf_height);
				z_draw_state->colors *= 2.0f;
			}
		}

		// reverse color change from DRAW_INTENSIFY workaround
		if(work_inten) z_draw_state->colors *= 2.0f;
	}


	// report our success or failure to do the actual drawing:
	//if(FAILED(status))
	//{
	//	z_Direct3DError(status);
	//	z_DirectDrawError(status);
	//	return 0;
	//}
	//else
	//{
		return 1;
	//}
}

// the simplest image drawing function -- using offscreen surfaces, usually very fast
// includes an option for masked drawing too, just input nomask = 0
int GameXClass::z_InternalDrawImageDD (LPDIRECTDRAWSURFACE7 p_sourceSurface, LPDIRECTDRAWSURFACE7 p_destSurface, RECT & src_rect, RECT & dst_rect, int nomask)
{
	// this function assumes:
	// both source and destination have already been UNlocked (if not, will refuse to draw anything)
	// the destination has a valid clipper attached to it (if not, won't clip and might crash)
	// (if the source and destination rectangles differ, Blt() will automatically scale the image,
	//	using hardware acceleration which will probably instantly scale it with bicubic filtering or something)

	z_EmptyVBuffer();
	z_LockVBuffer();

	int options = DDBLT_WAIT;
	if(!nomask) options |= DDBLT_KEYSRC;

	HRESULT status = p_destSurface->Blt(&dst_rect, p_sourceSurface,
										&src_rect,
										options,
										NULL);
	if(FAILED(status))
	{
		z_DirectDrawError(status);
		return 0;
	}
	return 1;
}


RECT GameXClass::z_ConvertRECT(RECT & rect, const ImageX * img)
{
	// somewhat kludgy way of getting a source rectangle:
	RECT rect2;
	if(rect.right != DEFAULT)
	{
		rect2 = rect;

		if(img != VIEWPORT && img != NULL)
		{
			if(img->z_xmult != 1.0f) rect2.left = (LONG)((float)rect2.left * img->z_xmult), rect2.right = (LONG)((float)rect2.right * img->z_xmult);
			if(img->z_ymult != 1.0f) rect2.top = (LONG)((float)rect2.top * img->z_ymult), rect2.bottom = (LONG)((float)rect2.bottom * img->z_ymult);
		}
	}
	else
	{
		if(img == VIEWPORT)
		{
			rect2.left = z_win_viewx1;
			rect2.top = z_win_viewy1;
			rect2.right = z_win_viewx2;
			rect2.bottom = z_win_viewy2;
		}
		else if(img == NULL)
		{
			rect2.left = 0;
			rect2.top = 0;
			rect2.right = 1;
			rect2.bottom = 1;
		}
		else
		{
			rect2.left = 0;
			rect2.top = 0;
			rect2.right = img->z_xres;
			rect2.bottom = img->z_yres;
			if(img->z_xmult != 1.0f) rect2.right = (LONG)((float)rect2.right * img->z_xmult);
			if(img->z_ymult != 1.0f) rect2.bottom = (LONG)((float)rect2.bottom * img->z_ymult);
		}
	}
	return rect2;
}

bool GameXClass::CopyToClipboard(const ImageX * img)
{
	LPDIRECTDRAWSURFACE7 p_src_surface;

	if(img == VIEWPORT)
	{
		p_src_surface = z_dd_back_buf;
	}
	else
	{
		if(img == NULL || img->z_status != IMG_OK || !img->z_hassurface || img->z_surface == NULL)
			return false;
		p_src_surface = img->z_surface;
	}

	return z_SurfaceToClipboard(z_dd_back_buf);
}

bool GameXClass::z_SurfaceToClipboard(const LPDIRECTDRAWSURFACE7 surface)
{
	bool ok = true;

	HDC hdc;
	DDSURFACEDESC2 ddsd;
	surface->GetDC (&hdc);
	HDC hdc2 = CreateCompatibleDC (hdc); 
	ZeroMemory (&ddsd, sizeof(ddsd));
	ddsd.dwSize = sizeof(ddsd);
	surface->GetSurfaceDesc (&ddsd);
	HBITMAP bitmap = CreateCompatibleBitmap (hdc, ddsd.dwWidth, ddsd.dwHeight);
	HBITMAP oldBitmap = (HBITMAP) SelectObject (hdc2, bitmap);
	BitBlt (hdc2, 0, 0, ddsd.dwWidth, ddsd.dwHeight, hdc, 0, 0, SRCCOPY);

	if(OpenClipboard (z_win_hwnd)) { 
		EmptyClipboard();
		SetClipboardData (CF_BITMAP,bitmap);
		CloseClipboard(); 
	} else
		ok = false;
	
	SelectObject (hdc2,oldBitmap);
	DeleteDC (hdc2);
	surface->ReleaseDC (hdc);
	return ok;
}

// will be called quite often, and so is optimized to take fast path in common case of things staying the same
void GameXClass::z_DetermineStates (LPDIRECTDRAWSURFACE7 image)
{
	if(image != z_last_texture_surface)
	{
		alphatype = 0;
		if(z_draw_state->src != VIEWPORT && z_draw_state->src != NULL)
		{
			if((z_draw_state->src->z_usage & LOAD_MASKED))
				alphatype = 1;
			else if((z_draw_state->src->z_usage & LOAD_ALPHA))
			{
				if(z_draw_state->dst == VIEWPORT || !(z_draw_state->dst->z_usage & LOAD_ALPHA))
				alphatype = 2; // alpha channel in image
			}
		}
		colorKeyEnable = (alphatype == 1) ? TRUE : FALSE; // masked status
	}

	static int lastFlags = -1;
	if(lastFlags != z_draw_state->flags || image != z_last_texture_surface)
	{
		lastFlags = z_draw_state->flags;

		// multiply by 2 if BRIGHT
		color_shift = 0;
		if(z_draw_state->flags & DRAWOP_BRIGHT)
			color_shift++;

		// apply draw mode workarounds, if necessary
		work_burn = false;
		work_inten = false;
		if(z_win_any_draw_mode_workaround)
			z_ApplyDrawModeWorkarounds(image, work_burn, work_inten);

		// determine src and dest blend modes from the draw mode
		z_ProcessDrawMode(srcBlend, destBlend, color_shift);

		invertEnable = (z_draw_state->flags & DRAWOP_INVERTED) ? D3DTA_TEXTURE|D3DTA_COMPLEMENT : D3DTA_TEXTURE; // if inverting, set to use the complement of the source texture
		cullEnable = (z_draw_state->flags & DRAWOP_NOCULL) ? D3DCULL_NONE : D3DCULL_CCW;
		filterEnable = (z_draw_state->flags & DRAWOP_NOFILTER) ? D3DTFG_POINT : D3DTFG_LINEAR;
		ditherEnable = (z_draw_state->flags & DRAWOP_NODITHER) ? FALSE : TRUE;
		alphaBlendEnable = (z_draw_state->flags & DRAWOP_NOBLENDALPHA && z_draw_state->flags & DRAW_PLAIN) ? FALSE : TRUE;
	}

	if(lastColorShift != color_shift || shade_combine_effect != z_draw_state->color_effect)
	{
		lastColorShift = color_shift;
		shade_combine_effect = z_draw_state->color_effect;

		if(color_shift > 0)
		{
			if(shade_combine_effect == D3DTOP_MODULATE)
			{
				if(color_shift == 1)
					shade_combine_effect = D3DTOP_MODULATE2X;
				else
					shade_combine_effect = D3DTOP_MODULATE4X;
			}
			else if(shade_combine_effect == D3DTOP_ADDSIGNED)
				shade_combine_effect = D3DTOP_ADDSIGNED2X;
		}
	}
}

void GameXClass::z_ApplyStates (LPDIRECTDRAWSURFACE7 image, LPDIRECT3DDEVICE7 device, bool forceSet)
{
	if(forceSet || lastSrcBlend != srcBlend)
	{
		device->SetRenderState(D3DRENDERSTATE_SRCBLEND,srcBlend);
		lastSrcBlend = srcBlend;
	}
	if(forceSet || lastDestBlend != destBlend)
	{
		device->SetRenderState(D3DRENDERSTATE_DESTBLEND,destBlend);
		lastDestBlend = destBlend;
	}
	if(forceSet || lastShadeCombineEffect != shade_combine_effect)
	{
		device->SetTextureStageState(0,D3DTSS_COLOROP, shade_combine_effect);
		lastShadeCombineEffect = shade_combine_effect;
	}
	if(forceSet || lastInvertEnable != invertEnable)
	{
		device->SetTextureStageState(0,D3DTSS_COLORARG1,invertEnable);
		lastInvertEnable = invertEnable;
	}
	if(forceSet || lastCullEnable != cullEnable)
	{
		z_d3d_device->SetRenderState(D3DRENDERSTATE_CULLMODE, cullEnable);
		lastCullEnable = cullEnable;
	}
	if(forceSet || lastColorKeyEnable != colorKeyEnable)
	{
		device->SetRenderState(D3DRENDERSTATE_COLORKEYENABLE,colorKeyEnable);
		lastColorKeyEnable = colorKeyEnable;
	}
	if(forceSet || lastFilterEnable != filterEnable)
	{
		device->SetTextureStageState(0,D3DTSS_MAGFILTER,filterEnable);
		device->SetTextureStageState(0,D3DTSS_MINFILTER,filterEnable);
		lastFilterEnable = filterEnable;
	}
	if(forceSet || lastDitherEnable != ditherEnable)
	{
		device->SetRenderState(D3DRENDERSTATE_DITHERENABLE, ditherEnable);
		lastDitherEnable = ditherEnable;
	}
	if(forceSet || lastAlphaBlendEnable != alphaBlendEnable)
	{
		device->SetRenderState(D3DRENDERSTATE_ALPHABLENDENABLE, alphaBlendEnable);
		lastAlphaBlendEnable = alphaBlendEnable;
	}
	if(forceSet || image != z_last_texture_surface)
	{
		device->SetTexture(0,image); // sets the texture image to draw
		z_last_texture_surface = image; // note: image NULL means blank/untextured poly
	}
}


		////**** Advanced Drawing Functions:

		//// functions for drawing meshes (still on the fly, no static data mesh support here)
		//void BeginDrawingVertices ();
		//inline void DrawVertex2D (float x, float y, DWORD c);
		//inline void DrawVertex2D (float x, float y, const ColorX & c);
		//inline void DrawVertex2D (float x, float y, float tu, float tv);
		//inline void DrawVertex2D (float x, float y, float tu, float tv, DWORD c);
		//inline void DrawVertex2D (float x, float y, float tu, float tv, const ColorX & c);
		//inline void DrawVertex3D (const Vec3f & v, DWORD c);
		//inline void DrawVertex3D (const Vec3f & v, const ColorX & c);
		//inline void DrawVertex3D (const Vec3f & v, float tu, float tv);
		//inline void DrawVertex3D (const Vec3f & v, float tu, float tv, DWORD c);
		//inline void DrawVertex3D (const Vec3f & v, float tu, float tv, const ColorX & c);
		//inline bool HasLastVertex (); // returns whether there is a last vertex available to repeat
		//inline void RepeatLastVertexPos (); // draws the last vertex (xy position only) that was drawn again - will fail if HasLastVertex() would return false
		//inline void RepeatLastVertexFull (); // draws the last vertex that was drawn again - will fail if HasLastVertex() would return false
		//void EndDrawingVertices (const ImageX* texture = NULL); // leave texture == NULL if drawing untextured

//void GameXClass::BeginDrawingVertices ()
//{
//	z_EmptyVBuffer();
//	z_LockVBuffer();
//}
//void GameXClass::EndDrawingVertices (const ImageX* texture)
//{
//	LPDIRECTDRAWSURFACE7 image = texture ? (texture == VIEWPORT ? z_dd_back_buf : (texture->z_hassurface ? texture->z_surface : NULL)) : NULL;
//	z_draw_state->src = texture;
//	z_DetermineStates(image);
///*
//	bool somethingChanged = 
//		   image != z_last_texture_surface
//		|| lastSrcBlend != srcBlend
//		|| lastDestBlend != destBlend 
//		|| lastCullEnable != cullEnable 
//		|| lastInvertEnable != invertEnable 
//		|| lastFilterEnable != filterEnable 
//		|| lastDitherEnable != ditherEnable 
//		|| lastColorKeyEnable != colorKeyEnable 
//		|| lastAlphaBlendEnable != alphaBlendEnable 
//		|| lastShadeCombineEffect != shade_combine_effect;
//
//	if((somethingChanged || z_d3d_vbufferfilled+6 >= z_d3d_vbuffermaxverts)
//	&& (z_d3d_vbuffer != NULL && z_d3d_vbuffermem != NULL && z_draw_state->dst == VIEWPORT && z_win_status == GAMEX_READY))*/
//	{
//		z_ApplyStates(image, z_d3d_device);
//
//		z_EmptyVBuffer();
//		z_LockVBuffer();
//	}
//}
//
//
//	inline bool GameXClass::HasLastVertex ()
//	{
//		return (z_d3d_vbufferfilled != 0);
//	}
//
//	inline void GameXClass::RepeatLastVertexPos ()
//	{
//		if(z_d3d_vbufferfilled != 0)
//		{
//			(*(z_d3d_vbuffermem)).sx = (*(z_d3d_vbuffermem-1)).sx;
//			(*(z_d3d_vbuffermem)).sy = (*(z_d3d_vbuffermem-1)).sy;
//			z_d3d_vbuffermem++;
//			z_d3d_vbufferfilled++;
//		}
//		else assert(!"RepeatLastVertexPos called with no previous vertex - should check HasLastVertex() first.");
//	}
//	inline void GameXClass::RepeatLastVertexFull ()
//	{
//		if(z_d3d_vbufferfilled != 0)
//		{
//			*(z_d3d_vbuffermem) = *(z_d3d_vbuffermem-1);
//			z_d3d_vbuffermem++;
//			z_d3d_vbufferfilled++;
//		}
//		else assert(!"RepeatLastVertexFull called with no previous vertex - should check HasLastVertex() first.");
//	}
//
//	inline void GameXClass::DrawVertex2D (float x, float y, DWORD c)
//	{
//		FVFTLVERTEX & vertex = *z_d3d_vbuffermem;
//
//		vertex.sx = x;
//		vertex.sy = y;
//		vertex.sz = z_draw_state->depth2D;
//		vertex.color = c;
//
//		z_d3d_vbuffermem++;
//		z_d3d_vbufferfilled++;
//	}
//	inline void GameXClass::DrawVertex2D (float x, float y, const ColorX & c)
//	{
//		FVFTLVERTEX & vertex = *z_d3d_vbuffermem;
//
//		vertex.sx = x;
//		vertex.sy = y;
//		vertex.sz = z_draw_state->depth2D;
//		vertex.color = RGBA_MAKE(c.c[0][0], c.c[0][1], c.c[0][2], c.c[0][3]);
//
//		z_d3d_vbuffermem++;
//		z_d3d_vbufferfilled++;
//	}
//	inline void GameXClass::DrawVertex2D (float x, float y, float tu, float tv)
//	{
//		FVFTLVERTEX & vertex = *z_d3d_vbuffermem;
//
//		vertex.sx = x;
//		vertex.sy = y;
//		vertex.sz = z_draw_state->depth2D;
//		vertex.rhw = 1.0f;
//		vertex.color = RGBA_MAKE(255,255,255,255);
//		vertex.tu = tu;
//		vertex.tv = tv;
//
//		z_d3d_vbuffermem++;
//		z_d3d_vbufferfilled++;
//	}
//	inline void GameXClass::DrawVertex2D (float x, float y, float tu, float tv, DWORD c)
//	{
//		FVFTLVERTEX & vertex = *z_d3d_vbuffermem;
//
//		vertex.sx = x;
//		vertex.sy = y;
//		vertex.sz = z_draw_state->depth2D;
//		vertex.color = c;
//		vertex.tu = tu;
//		vertex.tv = tv;
//
//		z_d3d_vbuffermem++;
//		z_d3d_vbufferfilled++;
//	}
//	inline void GameXClass::DrawVertex2D (float x, float y, float tu, float tv, const ColorX & c)
//	{
//		FVFTLVERTEX & vertex = *z_d3d_vbuffermem;
//
//		vertex.sx = x;
//		vertex.sy = y;
//		vertex.sz = z_draw_state->depth2D;
//		vertex.color = RGBA_MAKE(c.c[0][0], c.c[0][1], c.c[0][2], c.c[0][3]);
//		vertex.tu = tu;
//		vertex.tv = tv;
//
//		z_d3d_vbuffermem++;
//		z_d3d_vbufferfilled++;
//	}
//
//	inline void GameXClass::DrawVertex3D (const Vec3f & v, DWORD c)
//	{
//		FVFTLVERTEX & vertex = *z_d3d_vbuffermem;
//
//		z_draw_state->cam->Project(v, vertex.sx, vertex.sy, vertex.sz);
//		vertex.color = c;
//
//		const float &near_plane = z_draw_state->cam->GetNear(), &z_plane_sep = z_draw_state->cam->GetInvPlaneSep();
//		vertex.sz = (vertex.sz - near_plane) * z_plane_sep;
//
//		z_d3d_vbuffermem++;
//		z_d3d_vbufferfilled++;
//	}
//	inline void GameXClass::DrawVertex3D (const Vec3f & v, const ColorX & c)
//	{
//		FVFTLVERTEX & vertex = *z_d3d_vbuffermem;
//
//		z_draw_state->cam->Project(v, vertex.sx, vertex.sy, vertex.sz);
//		vertex.color = RGBA_MAKE(c.c[0][0], c.c[0][1], c.c[0][2], c.c[0][3]);
//
//		const float &near_plane = z_draw_state->cam->GetNear(), &z_plane_sep = z_draw_state->cam->GetInvPlaneSep();
//		vertex.sz = (vertex.sz - near_plane) * z_plane_sep;
//
//		z_d3d_vbuffermem++;
//		z_d3d_vbufferfilled++;
//	}
//	inline void GameXClass::DrawVertex3D (const Vec3f & v, float tu, float tv)
//	{
//		FVFTLVERTEX & vertex = *z_d3d_vbuffermem;
//
//		z_draw_state->cam->Project(v, vertex.sx, vertex.sy, vertex.sz);
//		vertex.rhw = 1.0f / vertex.sz;
//		vertex.color = RGBA_MAKE(255,255,255,255);
//		vertex.tu = tu;
//		vertex.tv = tv;
//
//		const float &near_plane = z_draw_state->cam->GetNear(), &z_plane_sep = z_draw_state->cam->GetInvPlaneSep();
//		vertex.sz = (vertex.sz - near_plane) * z_plane_sep;
//
//		z_d3d_vbuffermem++;
//		z_d3d_vbufferfilled++;
//	}
//	inline void GameXClass::DrawVertex3D (const Vec3f & v, float tu, float tv, DWORD c)
//	{
//		FVFTLVERTEX & vertex = *z_d3d_vbuffermem;
//
//		z_draw_state->cam->Project(v, vertex.sx, vertex.sy, vertex.sz);
//		vertex.rhw = 1.0f / vertex.sz;
//		vertex.color = c;
//		vertex.tu = tu;
//		vertex.tv = tv;
//
//		const float &near_plane = z_draw_state->cam->GetNear(), &z_plane_sep = z_draw_state->cam->GetInvPlaneSep();
//		vertex.sz = (vertex.sz - near_plane) * z_plane_sep;
//
//		z_d3d_vbuffermem++;
//		z_d3d_vbufferfilled++;
//	}
//	inline void GameXClass::DrawVertex3D (const Vec3f & v, float tu, float tv, const ColorX & c)
//	{
//		FVFTLVERTEX & vertex = *z_d3d_vbuffermem;
//
//		z_draw_state->cam->Project(v, vertex.sx, vertex.sy, vertex.sz);
//		vertex.rhw = 1.0f / vertex.sz;
//		vertex.color = RGBA_MAKE(c.c[0][0], c.c[0][1], c.c[0][2], c.c[0][3]);
//		vertex.tu = tu;
//		vertex.tv = tv;
//
//		const float &near_plane = z_draw_state->cam->GetNear(), &z_plane_sep = z_draw_state->cam->GetInvPlaneSep();
//		vertex.sz = (vertex.sz - near_plane) * z_plane_sep;
//
//		z_d3d_vbuffermem++;
//		z_d3d_vbufferfilled++;
//	}

bool GameXClass::z_IsBackBufLost()
{
	return z_dd_back_buf->IsLost() ? true : false;
}

HRESULT GameXClass::z_EndRenderBatchScene()
{
	return z_draw_render_batch_device->EndScene();
}


void GameXClass::ResetDrawStates()
{
	z_draw_state->flags = 0;
	SetDrawPart(0,0,DEFAULT,DEFAULT);
	z_draw_state->angle = 0.0f;
	z_draw_state->scalex = z_draw_state->scaley = 1.0f;
	z_draw_state->transx = z_draw_state->transy = 0.0f;
	z_draw_state->colors = ColorX();
	SetDrawShadingEffect(DRAW_MULTIPLY);
	z_draw_state->warp_mode = 0;
	z_draw_state->dst = VIEWPORT;
}

void GameXClass::SetDrawDepth (float depth)
{
	z_draw_state->depth2D = depth + z_draw_state->cam->GetNear();
}

void GameXClass::EnableAntialiasing ()
{
	z_d3d_device->SetRenderState(D3DRENDERSTATE_ANTIALIAS, TRUE);
}

void GameXClass::z_DrawDefaultCursor ()
{
	if(z_win_cursor_hidden && /*!z_CanHideCursor() &&*/ z_defaultCursorImg)
	{
		ResetDrawStates();
		GameX.SetDrawMode(DRAWOP_NOFILTER);
		DrawImage(z_defaultCursorImg, (float)GetMouseX(), (float)GetMouseY());
	}
}
