// GameX - GameXClass Class Header 
// 
// Copyright (C) 2002 Rama C. Hoetzlein
// 
// This software is released under the GameX GNU GPL 
// Open Source License. See the GameX documentation included
// with this source code for terms of modification, 
// distribution and re-release. 


#ifndef GAMEX_WIN_DX_H_INCLUDED
#define GAMEX_WIN_DX_H_INCLUDED

#define GAMEX_VERSION_STRING "GameX 5.09"

#include "gamex-global-def.h"

class GameXClass
{
public:
	GameXClass ();
	~GameXClass ();

	//**** Starting or Exiting GameX

	// Initializes GameX - must be called before using any graphical or audio parts of GameX.
	bool Initialize (const char* name = "Untitled Game", const GameXInitFlags options = VIDEO_ALLOWREFRESHSYNC, int xres = 1024, int yres = 768, int speed = 60);

	// Exits program, may ask for user confirmation first unless RUN_NOCONFIRMQUIT
	// is part of the Initialize options or unrecoverable is set to true.
	void Quit (bool unrecoverable = false);

	// Returns a (possibly empty) vector of all commandline arguments that were given to the program.
	const vector<const char*> & GetCommandLineArguments ();

	//**** Debugging Functions

	// Gets the global debugger, which you can use to output messages, open files, etc.
	// Note that you can also skip this step and just type things like debug.Output ("bla\n"); directly.
	inline Debugging GetDebug () {return debug;}

	// Automatically called by GameX if RUN_AUTOSHOWINFO is part of the Initialize option flags.
	// Draws the game cps,fps,idle status to the corner of the screen.
	void DrawStatus (bool continuously = false);

	//**** Interface Functions

	// Functions that allow you to easily display various types of dialog boxes (note: printf style formatting works)
	void InfoDialog (const char * messageWithFormat, ...); // gives info, user hits OK
	void WarningDialog (const char * messageWithFormat, ...); // gives warning info, user hits OK
	void ErrorDialog (const char * messageWithFormat, ...); // gives error info, user hits OK
	int ChoiceDialog (const char* message, bool three_choices = false, int default_choice = 1); // gives info, user hits YES, NO, or (if three_choices is true) CANCEL -- returns 0 for NO, 1 for YES, or -1 for CANCEL
	// see also: assert (), assertMsg (), and debug.AssertBox () in "gamex-debug.h"

	// Lets you set the icon that will appear in the taskbar -- must be called BEFORE GameX.Initialize ()
	// The input to this function should be an icon ID (such as IDI_ICON1) defined in a resource.h you include
	// which matches up with an icon resource in a .rc resource file you also include in your game project
	// (If you choose Project->Add Resource... in VS.NET and add an Icon resource to it, this is all created for you)
	void SetApplicationIcon (int iz_di_icon_id);


	//**** Random Number Functions (NOTE: these are independent in GameRun and GameDraw so as to be network/playback safe)

	// returns an int between 0 (inclusive) and 32767 (inclusive)
	short GetRand ();

	// returns an int between 0 (inclusive) and limit (exclusive)
	inline short GetRand (short limit) { return GetRand() % limit; }

	// returns an int between minim (inclusive) and maxim (inclusive)
	inline short GetRandomInt (short minim, short maxim) { return minim + (GetRand() % (maxim - minim + 1)); }

	// returns a float between minim (inclusive) and maxim (inclusive)
	inline float GetRandomFloat (float minim, float maxim) { return minim + ((float)GetRand() * (maxim - minim) / (float)RAND_MAX); }

	// setting the random seed - for advanced purposes involving reproduction of random number series
	void RandomizeSeed ();
	void SetRandomSeed (int seed);


	//**** Timing and Time-Related Functions

	// Gets a value that increases linearly with time
	// (the amount of game time that has gone by so far, in number of distinct game states.)
	// You can keep track of a variable for this on your own; this method is just for convenience.
	inline int GetGlobalCounter () {return z_win_global_counter;}
	inline void PostponeGlobalCounter () {z_win_global_counter--;} // call each cycle to pause the global counter

	// Gets a value that oscillates sinusoidally between a min and max value at a given speed.
	// Usage Example: float brightness = GameX.GetOscillatingValue (0.0f,255.0f,4.0f);
	// Note: phase is in DEGREES. so (for example) phase of 90 makes cosinusoidal oscillation
//	inline float GetOscillatingValue (float minim, float maxim, float speed, float phase = 0.0f) {return minim + (maxim-minim)/0.5f * (1.0f + sinf ( (phase+ (float)z_win_global_counter*speed)*DEGtoRAD) );}
template <typename T>
	inline T GetOscillatingValue (T minim, T maxim, float speed = 2.0f, float phase = 0.0f) { return (T)(minim + ((maxim - minim) * (0.5f * (1.0f + sinf((phase + (float)z_win_global_counter * speed) * DEGtoRAD))))); }
template <typename T>
	inline T GetCustomOscillatingValue (float counter, T minim, T maxim, float speed = 2.0f, float phase = 0.0f) { return (T)(minim + ((maxim - minim) * (0.5f * (1.0f + sinf((phase + counter * speed) * DEGtoRAD))))); }

	// Convert between a number of game cycles and the number of seconds it should take
	inline float CyclesToSeconds (int cycles) {return (float)cycles/ (float)z_win_cps;}
	inline int SecondsToCycles (float seconds) {return (int) (seconds* (float)z_win_cps);}


	//**** Performance Checking Functions

	// Gets the approximate percentage of CPU power that the game is currently NOT using.
	// Higher values mean the game is putting less strain on the CPU (which is good).
	inline int GetCurrentIdlePercent () {return z_p_idle;}

	// Gets the main loop/cycle rate the game is currently achieving, (calls/second of GameRun ())
	// Higher values returned means the game is running faster.
	inline int GetCurrentCycleRate ()	{return z_c_rate;}
	inline int GetTargetCycleRate ()	{return z_win_cps;}

	// Gets the frame rate the game is currently achieving, (calls/second of GameDraw ())
	// Higher values are better and mean the game is running smoother.
	inline int GetCurrentFrameRate ()	{return z_f_rate;}
	inline int GetTargetFrameRate ()	{return min(z_win_fps,z_win_cps);}


	//**** Altering Timing (optional functions to change game's auto-timing while it's running)

	void SetGameSpeed (int cps); // for dynamically changing the game speed that's initially set in GameX.Initialize ()
	void SetLagLimit (int max_cpf); // most lagged frames allowed per instant, or 0 to mean infinite (for networked games)
	void SetForcedFrameRate (int fps); // tries to force a certain framerate, for debugging use only
	void ResetForcedFrameRate (); // go back to targeting the best possible framerate that's visible
	inline int GetGameSpeed () {return GetCurrentCycleRate ();} // alternate name for GetCurrentCycleRate, to be used with SetGameSpeed
	

	//**** Game Screen Functions

	// Sets the viewport range, i.e. changes where GameX thinks the window is, for clipping
	// Games can use this for things like split-screen mode, along with CameraX::SetWindow ()
	void SetView (int xLeft, int yTop, int xRight, int yBottom);
	void GetView (int& xLeft, int& yTop, int& xRight, int& yBottom);
	void ResetView (); // resets view to entire window

	// Gets the current game screen width or height. It's better to use these than to hard-code.
	inline int GetWidth () {return z_win_width;}
	inline int GetHeight () {return z_win_height;}
	inline Vec2f GetScreenSize () {return Vec2f(z_win_width, z_win_height);}
	inline Vec2f GetScreenCenter () {return GetScreenSize() * 0.5f;}

	// Returns true if GameX is currently in full-screen mode, false otherwise
	inline bool IsFullScreen () {return z_win_fullscreen;}

	// Returns true if GameX is in "soft" (non-exclusive) full-screen mode
	inline bool IsFullScreenSoft () {return z_win_fullscreen && z_win_fake_fullscreen;}

	// Returns true if the game window is in the foreground
	bool IsWindowActive ();

	// gets the GameX window, for use with Windows functions
	inline HWND GetWindow () {return z_win_hwnd;}	
	inline int GetBPP () {return z_win_screenbpp;}

	// Gets the name the game was given when GameX.Initialize was called
	inline const char* GetName () 			{return z_win_name;}	

	
	//**** Drawing Functions: (for drawing sprites, shapes, polygons, backgrounds, etc.)
	// You can create a ColorX "on the fly" like so: GameX.FillScreen (ColorX (255,0,0));
	// But ImageX objects must be prepared first with ImageX::Create () or ImageX::Load ()

	// 2D drawing functions:
	void DrawPoint (const ColorX& clr, float x, float y);
	void DrawLine (const ColorX& clr, float x1,float y1, float x2,float y2);
	void DrawRect (const ColorX& clr, float xLeft, float yTop, float xRight, float yBottom);
	void DrawPolygon (const ColorX& clr, float x1,float y1, float x2,float y2, float x3,float y3, float x4,float y4);
	// vector versions
	inline void DrawPoint (const ColorX& clr, const Vec2f& v) { DrawPoint(clr,v.x,v.y); }
	inline void DrawLine (const ColorX& clr, const Vec2f& v1, const Vec2f& v2) { DrawLine(clr,v1.x,v1.y,v2.x,v2.y); }
	inline void DrawLine (const ColorX& clr, const LineSeg2f& line) { DrawLine(clr,line.p1.x,line.p1.y,line.p2.x,line.p2.y); }
	inline void DrawRect (const ColorX& clr, const Vec2f& topLeft, const Vec2f& topRight) { DrawRect(clr,topLeft.x,topLeft.y,topRight.x,topRight.y); }
	inline void DrawPolygon (const ColorX& clr, const Vec2f& v1, const Vec2f& v2, const Vec2f& v3, const Vec2f& v4) { DrawPolygon(clr,v1.x,v1.y,v2.x,v2.y,v3.x,v3.y,v4.x,v4.y); }

	// 2D image drawing functions:
	void DrawImage (const ImageX * img, float x, float y);
	void DrawImage (const ImageX * img, float x, float y, float angle, float scale); // shorthand for calling SetDrawAngle, SetDrawScale before drawing DrawImage
	void DrawLineImage (const ImageX * img, float x1,float y1, float x2,float y2, bool pixelUnitScale = false);
	void DrawTexturedRect (const ImageX * img, float xLeft, float yTop, float xRight, float yBottom);
	void DrawTexturedPolygon (const ImageX * img, float x1,float y1, float x2,float y2,  float x3,float y3,  float x4,float y4);
	// vector versions
	inline void DrawImage (const ImageX * img, const Vec2f& v) { DrawImage(img,v.x,v.y); }
	inline void DrawImage (const ImageX * img, const Vec2f& v, float angle, float scale) { DrawImage(img,v.x,v.y,angle,scale); }
	inline void DrawLineImage (const ImageX * img, const Vec2f& v1, const Vec2f& v2, bool pixelUnitScale = false) { DrawLineImage(img,v1.x,v1.y,v2.x,v2.y,pixelUnitScale); }
	inline void DrawLineImage (const ImageX * img, const LineSeg2f& line, bool pixelUnitScale = false) { DrawLineImage(img,line.p1.x,line.p1.y,line.p2.x,line.p2.y,pixelUnitScale); }
	inline void DrawTexturedRect (const ImageX * img, const Vec2f& topLeft, const Vec2f& topRight) { DrawTexturedRect(img,topLeft.x,topLeft.y,topRight.x,topRight.y); }
	inline void DrawTexturedPolygon (const ImageX * img, const Vec2f& v1, const Vec2f& v2, const Vec2f& v3, const Vec2f& v4) { DrawTexturedPolygon(img,v1.x,v1.y,v2.x,v2.y,v3.x,v3.y,v4.x,v4.y); }

	// 3D drawing functions:
	void DrawPoint3D (const ColorX& clr, const Vec3f& v);
	void DrawLine3D (const ColorX& clr, const Vec3f& v1, const Vec3f& v2);
	void DrawPolygon3D (const ColorX& clr, const Vec3f& v1, const Vec3f& v2, const Vec3f& v3, const Vec3f& v4);

	// 3D image drawing functions:
	void DrawImage3D (const ImageX * img, const Vec3f& v, bool worldUnitScale = true);
	void DrawLineImage3D (const ImageX * img, const Vec3f& v1, const Vec3f& v2, bool worldUnitScale = false);
	void DrawTexturedPolygon3D (const ImageX * img, const Vec3f& v1, const Vec3f& v2, const Vec3f& v3, const Vec3f& v4);

	// Screen-effect drawing functions:
	void ClearScreen ();
	void FillScreen (const ColorX& clr);
	void DrawTiled (const ImageX * img, float tx = 0, float ty = 0, float rx = DEFAULT, float ry = DEFAULT); // draws image tiled over screen, good for backgrounds or foregrounds
	inline void DrawTiled (const ImageX * img, Vec2f offset, Vec2f rotExtra = Vec2f(DEFAULT,DEFAULT)) { DrawTiled(img,offset.x,offset.y,rotExtra.x,rotExtra.y); }

	// Text Drawing Functions (same header as before for backward-compatibility):
	void DrawText (int x, int y, const char* msg, int r = 255, int g = 255, int b = 255, float fontSize = 24);
	void DrawOutlinedText (int x, int y, const char* msg, int r = 255, int g = 255, int b = 255, float fontSize = 24);
	void DrawEmbossedText (int x, int y, const char* msg, int r = 255, int g = 255, int b = 255, int xOff = 2, int yOff = 2, float fontSize = 24);

	// Advanced Text Drawing Functions:
	void DrawTextAdvanced (float x, float y, const char * msg, float fontSize, bool centered, int drawMode, ColorX & color, float angle = 0.0f, float yScale = 1.0f, bool rotateBaseline = true, bool forceFixedWidth = false);
	TextPrinter & GetTextPrinter () {return tp;}

	// GDI Text Drawing Functions (very slow and not very flexible):
	void DrawGDIText (int x, int y, const char* msg, int r = 255, int g = 255, int b = 255);
	void DrawOutlinedGDIText (int x, int y, const char* msg, int r = 255, int g = 255, int b = 255, int ro = 0, int go = 0, int bo = 0);

	// Special blurring functions (very slow, and may cause mosaic effect without graphic acceleration):
	void BlurScreen (float amount); // use sparingly
	void BlurRect (float amount, int xLeft, int yTop, int xRight, int yBottom);


	void SpecialPauseFade (float blur1, float blur2, int r1, int r2, int g1, int g2, int b1, int b2, int a);
	bool CopyToClipboard (const ImageX* img); // copies given ImageX (or VIEWPORT) to the Windows clipboard, returns false if failed, true otherwise

	bool ToggleFullscreen (bool immediate = true);
	bool ChangeGameSize (int xres, int yres);

	// optional states that can be set before calling the above drawing functions:

	inline void SetDrawCam (CameraX* cam) /* sets the 3D projection camera */ {z_draw_state->cam = cam;}
	inline void SetDrawMode (DrawFlags mode)  /* sets the drawing mode */ {z_draw_state->flags = mode;}
	inline void SetDrawPart (int xLeft, int yTop,	int xRight, int yBottom) /* sets to draw only a rectangular part of an image */ {z_draw_state->src_rect.left = xLeft; z_draw_state->src_rect.top = yTop; z_draw_state->src_rect.right = xRight; z_draw_state->src_rect.bottom = yBottom;}
	inline void SetDrawAngle (float angle)	/* sets drawing rotation angle in degrees counter-clockwise */ {z_draw_state->angle = angle;}
//		inline void SetDrawAngle (float angleX, float angleY, float angleZ) /* automatic 3D rotation of sprites not yet implemented */
	inline void SetDrawScale (float scale) /* sets drawing scale, can be a factor, # of pixels, or # of 3D world units, whichever makes sense */ {z_draw_state->scalex = scale; z_draw_state->scaley = scale;}
	inline void SetDrawScale (float scaleX, float scaleY) /* sets drawing scale separately for width and height */ {z_draw_state->scalex = scaleX; z_draw_state->scaley = scaleY;}
	inline void SetDrawTranslate (float transX, float transY) /* sets after-projection translation values, in pixels */ {z_draw_state->transx = transX; z_draw_state->transy = transY;}
	inline void SetDrawShading (const ColorX& color) /* set the image shading/transparency */ {z_draw_state->colors = color;}
	       void SetDrawShadingEffect (DrawFlags effect);
	inline void SetDrawWarp (float w0, float w1, float w2, float w3) {z_draw_state->warp_mode = 1; z_draw_state->warp[0] = w0; z_draw_state->warp[1] = w1; z_draw_state->warp[2] = w2; z_draw_state->warp[3] = w3;}
	       void SetDrawDestination (ImageX* dest);
	       void SetDrawDepth (float depth); // only affects 2D drawing into 3D scenes! -- is the distance in world units from the screen, 0.0 means as close as possible, 10.0f means further away, etc.
	       void ResetDrawStates ();


	//**** Getter method for certain states, for convenience
	inline DrawFlags GetDrawMode () {return z_draw_state->flags;}
	inline float GetDrawAngle () {return z_draw_state->angle;}
	inline float GetDrawScaleX () {return z_draw_state->scalex;}
	inline float GetDrawScaleY () {return z_draw_state->scaley;}
	inline float GetDrawTransX () {return z_draw_state->transx;}
	inline float GetDrawTransY () {return z_draw_state->transy;}
	inline CameraX* GetDrawCam () {return z_draw_state->cam;}
	inline ColorX& GetDrawShading () {return z_draw_state->colors;}
	inline DrawFlags GetDrawShadingEffect () {return z_draw_state->color_effect;}
	inline ImageX* GetDrawDestination () {return z_draw_state->dst;}

	//**** 3D Scene Functions

	// Tells GameX that you are about to draw a 3D scene to the viewport
	// so GameX will automagically draw the scene in the right order for you.
	// NOTE: A 3D scene means any bunch of things drawn with the DrawSomething3D functions,
	//       or 2D drawing functions when SetDrawDepth is used.
	void Begin3DScene ();

	// This is a hint to GameX that the things you draw after calling this function
	// don't need to be sorted relative to each other, and so can be drawn more quickly.
	// For instance, if you're drawing a bunch of additive particle effects,
	// draw them all last with a call to this function beforehand.
	inline void DoneSorting () {SetSortingMode (z_draw_sort_mode & SCENESORTZR);}

	// Tells GameX that you're done drawing a 3D scene.
	// Must be called immediately after drawing the scene or GameX may draw it wrong.
	// (note: this function clears the Z-buffer)
	void End3DScene ();


	// advanced use: if you know about z buffer read/write, transparent sorting,
	// and grouping additive calls nonsorted, this function gives you more control over that stuff
	void SetSortingMode (SceneSortMode sortMode);

	// advanced use: if you know you're switching draw modes or texture frequently,
	// disable attempts to strip together draw commands
	inline void DisableVertexBufferBatching () {z_d3d_vbufferenabled = false;}

	// advanced use: if you're drawing lots of things in a row with the same drawing mode and texture,
	// enable attempts to strip together draw commands
	inline void EnableVertexBufferBatching () {z_d3d_vbufferenabled = true; z_LockVBuffer();}

	// advanced use: enable automatic edge smoothing
	// WARNING: this will slow down drawing a LOT on most video cards
	//          and it is irreversible once called
	//          (and it will not significantly improve the appearance of most games)
	void EnableAntialiasing ();

	// advanced use: if you won't be using the z buffer at all, you can disable it for a small performance gain
	inline void DisableZBuffer () {if(z_d3d_zbuffer_enabled) z_d3d_zbuffer_enabled = false, z_d3d_zbuffer_wasEnabled = true;}
	inline void ReenableZBuffer () {if(z_d3d_zbuffer_wasEnabled) z_d3d_zbuffer_enabled = true, z_d3d_zbuffer_wasEnabled = false;}

	//**** Individual-Pixel-Drawing Functions:
	// WARNING: Not affected by clipping; game may crash if you draw a pixel out of bounds.

	// Tells GameX that you are about to read or write pixels
	// and if you pass it an ImageX it lets you set pixels in an image.
	// You must call this function before using DrawPixel or ReadPixel or they'll fail to work.
	void AccessPixels (ImageX* destination = VIEWPORT);
	void AccessPixels (const ImageX* destination = VIEWPORT);

	// Draws a pixel (must call AccessPixels first!) -- r,g,b range from 0 to 255
	inline void DrawPixel (int x,int y,int r,int g,int b,int a) { (*z_FuncDrawPixel) (x,y,r,g,b,a);}
	inline void DrawPixel (int x,int y,int r,int g,int b) { (*z_FuncDrawPixel) (x,y,r,g,b,255);}

	// Reads a pixel's values (must call AccessPixels first!) -- get r,g,b ranging from 0 to 255
	inline void ReadPixel (int x,int y, int &r, int &g, int &b, int &a) { (*z_FuncReadPixel) (x,y,r,g,b,a);}
	inline void ReadPixel (int x,int y, int &r, int &g, int &b) {int a; (*z_FuncReadPixel) (x,y,r,g,b,a);}

	// Tells GameX that you're done accessing individual pixels.
	// You MUST call this function before you can do any non-pixel drawing after calling AccessPixels.
	void EndPixelAccess ();


	//**** Sound Functions:
	// (must load a SoundX object with sound.LoadWav ("soundfile.wav"); before using the following)

	// Plays a sound at specified volume (0.0f to 1.0f), stereo panning (-1.0f to 0.0f to 1.0f -- left to middle to right)
	// and using a given frequency multiplier (1.0f for normal, 2.0f for double speed and pitch, 0.5f for half speed and pitch)
	void PlaySound (SoundX* snd, SoundPlayMode mode = PLAY_CONTINUE, float vol = 1.0f, float pan = 0.0f, float freq = 1.0f);

	// Stops a sound from playing if it was playing/looping
	void StopSound (SoundX* snd);

	// Sets the sound to play from its beginning
	void RewindSound (SoundX* snd);

	// Returns true if the given sound is still playing, looping, or paused, false otherwise
	bool IsSoundPlaying (const SoundX* snd);


	//**** Music Functions:
	// (must load a MusicX object with music.Load (music_filename); before using the following)

	// Plays a MusicX object, and immediately stops playing any other music that may be playing
	// times of 0 means infinite loop, 1 means play once, 2 means play twice, etc.
	// speed_factor of 1.0 means normal speed, 2.0 means double speed, etc.
	void PlayMusic (MusicX* music, int times = 0, float volume = 1.0f, float fade_in_seconds = 1.0f, float tempo = 1.0f, float pitch = 1.0f);
	inline void PlayMovie (MovieX* movie, int times = 0, float volume = 1.0f) { PlayMusic((MusicX*)movie, times, volume); }

	// Fades the volume to 0 over given seconds if seconds>0, then stops the music from playing
	void StopMusic (MusicX* music, float seconds = 2.0f);
	inline void StopMovie (MovieX* movie) { StopMusic((MusicX*)movie); }

	// Fades the music up or down to the given volume factor, over the given time
	// seconds is how many seconds to alter volume over
	// target_factor of 0.0f is fade to silence, 0.5f is fade to half master volume, 1.0f is no change, 1.27f is fade to max amplified
	void VolumeFadeMusic (MusicX* music, float volume_factor = 0.0f, float seconds = 0.0f);

	// Bends the pitch of the mucis up or down to the given pitch factor, over the given time
	// seconds is how many seconds to alter pitch over. This method does not change speed of play
	// pitch_factor of 0.5f is half pitch (low frequency), 2.0f is double pitch (high frequency), 1.0f is no change, etc.
	void PitchBendMusic (MusicX* music, float pitch_factor = 0.5f, float seconds = 0.0f);

	// Sets the tempo, or speed of play, of all music -- does not change the pitch
	// 0.5f is half speed, 1.0f is normal speed, 2.0f is double speed, etc.
	// Can be called before or after you start playing the song
	// note: there may be a small time delay after calling this function before the change takes effect
	void SetMusicTempo (float tempo_factor);

	// Returns whatever tempo was last set by GetMusicTempo, or the default of 1.0f
	float GetMusicTempo () {return z_master_music_tempo;}

	// Sets the volume of all music
	// Can be called before or after you start playing a song
	// If an individual music song has its volume faded to 50% and the master volume is 50%
	// that will result in the music being played at 25% volume
	void SetMasterMusicVolume (float volume_factor);

	// Returns whatever volume was last set by SetMasterMusicVolume, or the default of 1.0f
	float GetMasterMusicVolume () {return z_master_music_volume;}

	// returns true if the given song is still playing, looping, or paused, false otherwise
	bool IsMusicPlaying (const MusicX* music);
	inline bool IsMoviePlaying (const MovieX* movie) { return IsMusicPlaying((const MusicX*)movie); }

	// pauses whatever music/movies are currently playing (has only been tested internally to GameX)
	void PauseMusic ();

	// resumes whatever music/movies song are currently playing (has only been tested internally to GameX)
	// due to the way it is paused, for MIDI files a stray note might play for a while after resuming
	void ResumeMusic ();
	

	//**** Keyboard Input Functions:
	// NOTE: See gamex-defines.hpp for the key constants, like KEY_RETURN and KEY_SPACE

	// Old input functions, for backward compatibility and for developers who only need a simple interface:
	inline bool GetKeyDown (KeyID k)   {return IsKeyDown (k);}
	inline char GetKey ()		   {return GetBufferedKeyPress ();}

	// Returns the next char of buffered key input (like 'A' or 'b' or '3'), returns 0 if there is no key input left
	// Use this to get typed data input from the user (such as a name at a name entry screen)
	char GetBufferedKeyPress ();

	// Returns the next char of buffered key input without advancing past it, returns 0 if there is no key input left
	// Use this if you want to see what GetBufferedKeyPress would return without actually taking it out of the buffer
	char PeekBufferedKeyPress ();

	// Completely clears the key buffer
	// Use this, for instance, before asking for input to clear unwanted data from earlier
	void ClearKeyBuffer ();


	// Returns true if the given key is down, false otherwise
	// Use this if you only want to know if a given key is down or not (like the trigger of a machine gun)
	inline bool IsKeyDown (KeyID k) 	 {return (z_di_keystate[k] & INPUT_PRESSED) ? true : false;}

	// Returns true if the given key is down and was up at least once since the last time this function was called on it, false otherwise
	// Use this if you want to see if a key was just pressed (not just being held down, so you only respond once)
	inline bool GetKeyPress (KeyID k)		  {return (z_di_keypressed[k] & INPUT_PRESSED) ? true : false;}

	// Returns true if the given key is down and was not held down since the last time GetKeyPress was called on it, false otherwise
	// Use this if you want to see if a key was just pressed but don't want to also mark it as being held down yet
	inline bool PeekKeyPress (KeyID k)		  {return (z_di_keypressed[k] & INPUT_PRESSED) ? true : false;}
	
	// Returns true if any key on the keyboard is down and was up at least once since GetKeyPress was called on it, false otherwise
	// Use this at "Press Any Key" prompts. Note that this automatically excludes certain keys, such as volume control keys that a game should not respond to.
	bool AnyKeyPress ();
	bool PeekAnyKeyPress ();

	// Convenience methods for checking for modifier keys being held down:
	inline bool IsShiftDown ()	{return IsKeyDown (KEY_LSHIFT) || IsKeyDown (KEY_RSHIFT);}
	inline bool IsAltDown ()	{return IsKeyDown (KEY_LALT)   || IsKeyDown (KEY_RALT);}
	inline bool IsCtrlDown ()	{return IsKeyDown (KEY_LCTRL)  || IsKeyDown (KEY_RCTRL);}
	inline bool IsWinDown ()	{return IsKeyDown (KEY_LWIN)   || IsKeyDown (KEY_RWIN);}

	// To check whether caps/num/scroll lock is on:
	inline bool IsCapsLockOn ()   {return (GetKeyState (VK_CAPITAL) != 0);}
	inline bool IsNumLockOn ()	   {return (GetKeyState (VK_NUMLOCK) != 0);}
	inline bool IsScrollLockOn ()	{return (GetKeyState (VK_SCROLL) != 0);}

	// Non-polling key up/down event getters (for advanced use)
	// these return keys like KEY_LEFT, KEY_A, etc., or KEY_NOTHING for no key event.
	// these will not tell you whether a key is currently being held down, that's what IsKeyDown is for
	KeyID       GetKeyDownEvent ();
	KeyID       GetKeyUpEvent ();
	KeyID       PeekKeyDownEvent ();
	KeyID       PeekKeyUpEvent ();
	void        ClearKeyDownEvents ();
	void        ClearKeyUpEvents ();


	//**** Mouse Input Functions:
	// NOTE: See gamex-defines.hpp for the mouse constants, like MOUSE_LEFT and MOUSE_RIGHT 

	inline int GetMouseX () {return z_mouse_x*z_win_width/z_win_realx;} // mouse x position, right is positive (scaled to the virtual screen so the game can use it without worrying about size variations that it can't control)
	inline int GetMouseY () {return z_mouse_y*z_win_height/z_win_realy;} // mouse y position, down is positive (scaled to the virtual screen)
	inline int GetMouseDX () {return z_di_mousestate->lX;} // mouse x movement, right is positive
	inline int GetMouseDY () {return z_di_mousestate->lY;} // mouse y movement, down is positive
	inline int GetMouseDZ () {return z_di_mousestate->lZ;} // mouse wheel movement, up is positive

	// sets the mouse position in terms of game pixels, (0,0) being top-left corner of the game view
	// valid (x,y) are between (0,0) and (z_win_width-1,z_win_height-1) 
	void SetMousePosition (int x, int y);

	// returns true if the given mouse button is down, false otherwise
	inline bool IsMouseDown (MouseButtonID b)	 {return (z_di_mousestate->rgbButtons[b] & INPUT_PRESSED) ? true : false;}

	// returns true if the given mouse button is down and was up at least once since the last time this function was called on it, false otherwise
	inline bool GetMouseClick (MouseButtonID b)		  {return (z_di_mousepressed[b] & INPUT_PRESSED) ? true : false;}

	// returns true if the given mouse button is down and was not held down since the last time GetMouseClick was called on it, false otherwise
	// (i.e. returns true for as long as the mouse button is held down after a click, can be used for click-drags)
	inline bool PeekMouseClick (MouseButtonID b) 	  {return (z_di_mousepressed[b] & INPUT_PRESSED) ? true : false;}

	// returns true if any mouse button is down and was up at least once since GetMouseClick was called on it, false otherwise
	inline bool AnyMouseClick () 		{for(MouseButtonID b = 0; b < 4; b++) if(GetMouseClick (b)) return true; return false;}
	inline bool PeekAnyMouseClick () 	{for(MouseButtonID b = 0; b < 4; b++) if(PeekMouseClick (b)) return true; return false;}

	// returns true if the given mouse button is double-clicked and was up at least once since the last time this function was called on it, false otherwise
	// use this if you want to respond once to a double-click on a certain mouse button
	inline bool GetMouseDoubleClick (MouseButtonID b)  {if(z_di_mousepressed[b] & INPUT_DOUBLE) {z_di_mousepressed[b] ^= (INPUT_HELD|INPUT_DOUBLE); return true;} else return false;}

	// returns true if the given mouse button is double-clicked and was not held down since the last time GetMouseDoubleClick was called on it, false otherwise
	// (i.e. returns true for as long as the mouse button is held down after a double-click, can be used for double-click-drags)
	inline bool PeekMouseDoubleClick (MouseButtonID b) {return (z_di_mousepressed[b] & INPUT_DOUBLE) ? true : false;}


	// Non-polling mouse up/down event getters (for advanced use)
	// these return buttons like MOUSE_LEFT, MOUSE_RIGHT, etc., or MOUSE_NOTHING for no mouse button event
	// these will not tell you whether a button is currently being held down, that's what IsMouseDown is for
	MouseButtonID GetMouseDownEvent ();
	MouseButtonID GetMouseUpEvent ();
	MouseButtonID PeekMouseDownEvent ();
	MouseButtonID PeekMouseUpEvent ();
	void          ClearMouseDownEvents ();
	void          ClearMouseUpEvents ();



	// get displayable string names of key or mouse ID codes
	void GetKeyString (char * output, KeyID key);
	void GetMouseButtonString (char * output, MouseButtonID button);



	// Automatically called by GameX -- Copies what's been drawn so far to the screen. 
	// WARNING: The game should NEVER call this function except under certain special circumstances (i.e. drawing loading screen inside GameInit).
	int UpdateView ();

	// can use this to override the draw warning if you know what you're doing
	void SetWarningOnDraw (bool set) {z_win_warning_on_draw = set;}

	// public so the classless windows app can access
	int z_StartMain (HINSTANCE hThisInst, HINSTANCE hPrevInst, LPSTR lpszArgs, int nWinMode);
	LRESULT z_HandleMessage (HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam);
	HRESULT z_HandleDeviceEnum (LPSTR lpDeviceDescription, LPSTR lpDeviceName, LPD3DDEVICEDESC7 lpD3DDeviceDesc, LPVOID lpContext);
	HRESULT z_HandleZBufferFormatEnum (LPDDPIXELFORMAT lpDDPixFmt, LPVOID lpContext);
	HRESULT z_HandleDisplayEnum (LPDDSURFACEDESC2 lpDDModeDesc, LPVOID lpContext);
	int z_SafeMessageBox(const char* text, const char* title, unsigned int type);
	void z_ReportProblem (GameXStatus problem);
	inline GameXStatus z_GetStatus () 		   {return z_win_status;}
	inline void z_SetStatus (GameXStatus status)	   {z_win_status = status;}
	long long			z_sleep_record;
	int z_win_maskr,	z_win_maskg,	z_win_maskb;
	int z_win_shiftr, z_win_shiftg, z_win_shiftb;
	int z_win_maxr,	z_win_maxg,	z_win_maxb;
	int z_win_maskminr, z_win_maskming, z_win_maskminb; // minimum nonzero values out of 256 (like as 6 or 8)
	char*					z_win_pixel_dest_data;
	int*					z_win_pixel_dest_ytable;
	int 					z_win_pixel_dest_pitch;
	float					z_win_pixel_dest_xmult, z_win_pixel_dest_ymult;

protected:
	int 			z_win_ytable[4096+32]; // used (in pixel drawing) by GameX's static functions, for speed
	bool			z_win_active; // true if GameX window is in front and active
	int 			z_win_realx, z_win_realy; // width/height of visible window/fullscreen area, and bpp
	int 			z_win_global_counter, z_win_global_frame_total;
	int 			z_win_cps; // speed of game -- number of times GameRun () is called per second
	int 			z_win_fps; // smoothness of game -- number of times GameDraw () is called per second
	int 			z_win_max_cpf; // "jerkiness" limit -- max cycles that can be skipped without drawing -- set really high if you want no limit
	unsigned short	z_f_rate; // current frame rate so game can know how many fps it's getting
	unsigned short	z_c_rate; // current cycle rate so game can know how many cps it's getting
	int 			z_p_idle; // CPU time not uses (as a percent)
	bool			z_win_alternator;
	LPD3DDEVICEDESC7		z_d3d_device_desc;
	LPDDPIXELFORMAT			z_d3d_zbuffer_format;
	bool z_drawLocked;
	int 					z_win_xpos, z_win_ypos; 	
	int 					z_win_xoffset, z_win_yoffset;	



	bool					z_win_log_debug;
	bool					z_win_err_debug;

	bool					z_win_init_done;


private:
	// note: prefix z_ is to sort it below all the private variables in VS.NET listings - we could also use $ to hide it completely but that has some other weird side effects...
	inline LPDIRECTDRAW7 z_GetDD ()		{return z_dd_main;}
	inline LPDIRECTSOUND z_GetDS ()		{return z_ds_main;}
	inline IDirectMusicPerformance* z_GetDMP ()	{return z_dm_performance;}
	inline IDirectMusicLoader* z_GetDML ()		{return z_dm_loader;}
	inline bool z_SupportsDepth (int bpp) 	{return (bpp == 16 || bpp == 24 || bpp == 32);}
	inline bool z_CanAutoPause () 		{return (!z_win_request_no_auto_pause);}	
	inline bool z_CanHideCursor ()		{return (!z_win_request_no_hide_cursor /*&& (z_win_request_no_show_cursor)*/);}	
	inline bool z_CanUseMouse ()			{return (!z_win_request_no_mouse_input);} 
	inline bool z_CanShowRunInfo ()		{return (z_win_request_show_run_info);}	
	inline bool z_CanDeleteMData ()		{return (z_win_request_delete_mdata);}	
	inline bool z_CanDraw ()				{return (z_win_init_done && z_win_status == GAMEX_READY && z_dd_back_buf != NULL && !z_isdrawing && !z_IsBackBufLost());}	
	inline bool z_CanPlayMP3 ()			{return z_mp3_support_tested ? z_mp3_support : z_TestDirectShow ();}
	inline bool z_IsEscapeReserved () 	{return !z_win_request_use_escape;}
	inline bool z_IsAltEnterReserved ()	{return !z_win_request_use_alt_enter;}
	inline bool z_IsQuitDialogOpen () 	{return z_quit_dialog;}
	inline bool z_IsResizeWorkaroundEnabled () {return z_win_resize_workaround ? true : false;}
	inline bool z_SimulatingSoftwareOnly (){return z_win_request_software_only;}
	inline LPDIRECTDRAWSURFACE7 z_GetBackSurf () {return z_dd_back_buf;}
	bool z_CheckQuit ();
	void z_DisplayLoadingWithValue (int percent, const char * message = NULL);
	void z_MemoryRecordChange (int change) {z_memory_used_current += change; z_memory_used_max = max(z_memory_used_max,z_memory_used_current);}
	int z_GetGraphicMemUsed () {return z_memory_used_current;}
	void z_SetImageResolutionLimit (int size) {z_d3d_max_tex_width = size; z_d3d_max_tex_height = size;}
	bool z_InitDirectMusicPerformance ();
	bool z_InitDirectMusicLoader ();
	void z_AccessPixelsHelper (const ImageX* destination, bool readOnly);

	void z_HideCursor ();
	void z_UnhideCursor ();



	// locking / unlocking drawing surface
	bool z_LockBackBuffer ();
	void z_UnlockBackBuffer ();
	void z_HandleDirectShowEvent ();

	int z_GetShift (int mask);
	int z_GetMax (int mask);
	LPDIRECTDRAWSURFACE7 z_CreateImageSurface (int width, int height, bool texture, int create_alpha, bool targetable);
	bool z_CreateDeviceForSurface (LPDIRECT3DDEVICE7& device, LPDIRECTDRAWSURFACE7& surface);

	//*** Windows Control Functions
	// Report a Problem Initializing
	void z_ReportProblem ();		
	// Initialization & Clean Up Functions		
	void z_InitWindowsInstance (HINSTANCE hThisInst, int nWinMode);	 
	void z_ParseArgs (char* args);
	void z_DirectDrawError (HRESULT status);
	void z_Direct3DError (HRESULT status);
	void z_CleanUp ();		
	// Internal State Control Functions
	inline void z_SetActivation (bool act) {z_win_active = act;}
	inline bool z_GetActivation () {return z_win_active;}
	void z_SetPosition (int x, int y);		
	void z_SetProportional ();
	void z_SetMouseMove (int x, int y);
	void z_SetMouseStop (int disable_length) {z_di_mousestop = disable_length;} 

	// private input functions
	void z_AddBufferedKeyPress (char ch);
	void z_AddKeyDownEvent (int key);
	void z_AddKeyUpEvent (int key);
	void z_AddMouseDownEvent (MouseButtonID btn);
	void z_AddMouseUpEvent (MouseButtonID btn);
	void z_SetMouseRelease (int button)		{z_mouse_down[button] = 0;}
	void z_SetMousePress (int button)		{z_mouse_down[button] = 1;}
	void z_SetMouseDoubleClick (int button)	{z_mouse_down[button] = 2;}

	// used internally when a MusicX object is loaded:
	inline void z_RegisterMusicX (MusicX* mxp) {z_dm_musicx[z_dm_musicx_index] = mxp; if(z_dm_musicx_index<255) z_dm_musicx_index++;}
	inline void z_RegisterSoundX (SoundX* sxp) {z_sound_list.push_back (sxp);}
	inline void z_UnRegisterMusicX (MusicX* mxp) {FOR(i,255) if(z_dm_musicx[i] == mxp) z_dm_musicx[i] = NULL;}
	inline void z_UnRegisterSoundX (SoundX* sxp)
	{
		for(std::vector<SoundX*>::iterator iter = z_sound_list.begin () ; iter != z_sound_list.end () ; iter++)
		{
			if(*iter == sxp)
			{
				z_sound_list.erase (iter);
				break;
			}
		}
	}
	void z_UpdateSound (const SoundX* snd) ;

	bool z_UpdateInput ();					   // Retrieve input

	bool z_SurfaceToClipboard (const LPDIRECTDRAWSURFACE7 surface); // copies given surface to the clipboard



	void z_InternalDrawMaster (const ImageX* source); // master drawing function that most drawing functions use

	// Draw an image using Direct3D (hopefully 3D-accelerated)
	int z_InternalDrawImageD3D (LPDIRECT3DDEVICE7 device, LPDIRECTDRAWSURFACE7 image, RECT& src_rect, RECT& src_bounds, float* x, float* y, float* z, int surf_width, int surf_height);

	// Draw an image using DirectDraw (hopefully 2D-accelerated)
	int z_InternalDrawImageDD (LPDIRECTDRAWSURFACE7 p_src_surface, LPDIRECTDRAWSURFACE7 p_dst_surface, RECT& src_rect, RECT& dst_rect, int nomask);
	void z_InternalDrawRectDD (int red, int green, int blue, int xLeft, int yTop, int xRight, int yBottom); // fast but cannot use any advanced features like transparency or rotation

	// Clipping Functions:
	// source to destination (image/viewport) boundary clipping function:
	int z_ClipRects (RECT& src_rect, RECT& dst_rect, const RECT& src_bounds, const RECT& dst_bounds); // clips source and dest rects to their boundaries
	void z_ClipTile (int x, int y, int xres, int yres, int &mx, int &my, int &offx, int &offy);
	RECT z_ConvertRECT (RECT& rect, const ImageX * img);

	void z_NullDrawPixelFuncs ();

	inline long z_ToAttenuation (float vol) {return (long) (vol <= 0 ? -10000 : logf (vol)*1000.f);} // converts a value on a 0.0 to 1.0 volume scale to a logarithmic volume attenuation value

	// helper function for functions that want to convert 4 sets of (r,g,b,a) to (r,g,b)
	void z_ScaleByAlpha (ColorX& colors);

	inline HRESULT z_EndRenderBatch () {z_draw_render_batch_open = false; return z_draw_render_batch_device ? z_EndRenderBatchScene () : 0;}

	bool z_IsBackBufLost ();
	HRESULT z_EndRenderBatchScene ();



	// Dispatched Drawing Functions
	void (*z_FuncDrawPixel) (int x, int y, int r, int g, int b, int a);
	void (*z_FuncReadPixel) (int x, int y, int &r, int &g, int &b, int &a);

	// fills quadrilateral with screen vertices,
	// returns false if it isn't on the screen
	bool z_GetScreenVertices (float* x, float* y, float* z);

	bool z_GetScreenVertices (const ImageX* source, bool immediate = false);
	void z_ApplyDrawModeWorkarounds (LPDIRECTDRAWSURFACE7 image, bool & work_burn, bool & work_inten);
	void z_ProcessDrawMode (int & srcBlend, int & destBlend, int & color_shift);

	void z_EmptyVBuffer (); // draws and empties the vertex buffer
	void z_LockVBuffer (); // locks the vertex buffer, preparing it to be filled again

	void z_DetermineStates (LPDIRECTDRAWSURFACE7 image);
	void z_ApplyStates (LPDIRECTDRAWSURFACE7 image, LPDIRECT3DDEVICE7 device, bool forceSet = false);


	void z_SyncWithRefresh (); // Returns once the scanline is clear of the game window or screen

	void z_DoInitTests ();

	void z_LoadQueuedImages ();
	void z_DrawDefaultCursor ();

	// Windows & DirectX Specific Initialization
	bool z_InitWindowSetup ();
	bool z_InitDirectX ();
	bool z_InitWindowClass ();							// Create a window class
	bool z_InitWindows ();								 // Create a main window
	bool z_WindowDraw (); 								// Draw window on screen
	bool z_InitDirectDraw (); 							// Initialize direct draw
	bool z_InitDirect3D ();
	bool z_InitDirectInput ();
	bool z_InitDirectSound ();
	bool z_TestDirectShow ();
	void z_InitVSync ();
	bool z_ReInitScreen ();
	bool z_InitDirectDrawSurfaces (); 				// Initialize direct draw surfaces
	bool z_InitDirectDrawMonitor ();					// Initialize monitor settings for direct draw
	bool z_DirectXCleanUp (); 						// Cleanup direct draw



private:
	HINSTANCE				z_win_hinst;	
	int 					z_win_hmode;
	vector<const char*>		z_win_commandLineArgs;

	GameXStatus 			z_win_status; 
	int 					z_win_request_xsize, z_win_request_ysize; // requested width and height of view area
	int 					z_win_request_bpp; // VIDEO_16BIT makes this 16, VIDEO_32BIT makes this 32, etc., 0 means no request
	int 					z_win_request_full; // 0 if VIDEO_WINDOWED flag on, 1 if VIDEO_FULLSCREEN flag on, 2 if neither is on, temporarily
	int 					z_win_request_vsync; // 0 unless VIDEO_ALLOWREFRESHSYNC flag on
	bool					z_win_request_no_sound; // false unless AUDIO_DISABLE flag on
	bool					z_win_request_antialias; // false unless VIDEO_ALLOWANTIALIAS flag on
	bool					z_win_request_can_resize; // true unless VIDEO_NORESIZE flag on
	bool					z_win_request_use_escape; // false unless RUN_USEESCAPEKEY flag on
	bool					z_win_request_delete_mdata; // false unless RUN_ASSUMEVRAMSAFE flag on
	bool					z_win_request_use_alt_enter; // false unless NOALTENTERTOGGLE flag on
	bool					z_win_request_software_only; // false unless VIDEO_SOFTWAREONLY flag on
	bool					z_win_request_no_auto_pause; // false unless RUN_BACKGROUND flag on
	bool					z_win_request_show_run_info; // false unless RUN_AUTOSHOWINFO flag on
	bool					z_win_request_low_resolution; // false unless VIDEO_LOWRES flag on
	bool					z_win_request_no_hide_cursor; // false unless RUN_ALWAYSKEEPCURSOR flag on
	bool					z_win_request_no_show_cursor; // false unless RUN_ALWAYSHIDECURSOR flag on
	bool					z_win_request_no_mouse_input; // false unless RUN_NOMOUSEINPUT flag on
	bool					z_win_request_no_confirm_quit; // false unless RUN_NOCONFIRMQUIT flag on
	bool					z_win_request_soft_fullscreen; // false unless VIDEO_SOFTFULLSCREEN flag on
	bool					z_win_request_hard_fullscreen; // false unless VIDEO_HARDFULLSCREEN flag on
	bool					z_win_request_slow_debug_drawing; // false unless VIDEO_SLOWDEBUGDRAWING flag on
	bool					z_win_request_keep_cursor_in_window; // false unless RUN_KEEPCURSORINWINDOW flag on

	bool					z_win_status_problem_ignore [GAMEX_TOTAL_STATUS_NUMBER]; // true for each GameXStatus problem user ignores

	bool					z_win_fullscreen, z_win_fake_fullscreen; 
	bool					z_win_cursor_hidden;
	bool					z_win_vsync_enabled; // true if vsync is on and currently ready
	int 					z_win_size;
	int 					z_win_widbits;
	bool					z_win_back_buf_in_vid_mem;
	bool					z_win_warning_on_draw;

	HWND					z_win_hwnd;
	LPCTSTR					z_win_appIconIDStr;

	char					z_win_name [64];

	char*					z_win_screen;
	int 			z_win_height, z_win_width; // width/height of the virtual game screen
	int 			z_win_pitch, z_win_adjustedpitch; // pitch and bpp-adjusted pitch of the virtual game screen
	int 			z_win_screenx, z_win_screeny, z_win_screenbpp, z_win_screenrefresh; // width/height of visible window/fullscreen area, and bpp
	int 			z_win_viewx1, z_win_viewy1, z_win_viewx2, z_win_viewy2;

	bool					z_quit_dialog;

	ImageX*					z_defaultCursorImg;

	LPDIRECTDRAWSURFACE7	z_win_pixel_dest_surface;

	bool					z_draw_render_batch_open;
	LPDIRECT3DDEVICE7		z_draw_render_batch_device;

	int 					z_draw_3D_depth_sort; // 1 if sorting a 3D scene, 2 if actually drawing it, 0 if done or not doing either
	SceneSortMode			z_draw_sort_mode;
	GameXDrawState*			z_draw_state; // current draw state
	list<GameXDrawState>	z_draw_state_list;
	
	bool					z_win_blur_supported; // if blurring of entire game screen is fast enough to allow (should = true except on old video cards)
	bool					z_win_any_draw_mode_workaround; // if any drawmode related workarounds are on
	int 					z_win_subtract_workaround; // 0 if subtraction works, 1 if works for alpha images only, 2 if doesn't work and requires black-alpha subtraction simulation (should = 0 except on old video cards)
	int 					z_win_intensify_workaround; // 0 if intensification works, 1 if doesn't work and regular addition should be used instead
	int 					z_win_addsoft_workaround; // 0 if DRAW_ADDSOFT works, 1 if doesn't work and regular addition should be used instead
	int 					z_win_addsharp_workaround; // 0 if DRAW_ADDSHARP works, 1 if doesn't work and regular addition should be used instead
	int 					z_win_ghost_workaround; // 0 if DRAW_GHOST works, 1 if it requires a color-to-alpha copied workaround image drawn normally to work
	int 					z_win_burn_workaround; // 0 if DRAW_BURN works, 1 if it doesn't work and should instead DRAW_INVERT then non-color-shifted DRAW_SUBTRACT
	int 					z_win_alpha_blend_workaround; // 0 if alpha-blend shading works on non-alpha images, 1 otherwise
	int 					z_win_resize_workaround; // 1 if stretched blitting is too slow to allow window resizing to stretch the game screen, 0 otherwise

	std::queue<char>			z_win_keycharbuf;
	std::queue<KeyID>			z_win_keydownbuf;
	std::queue<KeyID>			z_win_keyupbuf;
	std::queue<MouseButtonID>	z_win_mousedownbuf;
	std::queue<MouseButtonID>	z_win_mouseupbuf;

	int 					z_mouse_x, z_mouse_y;
	int 					z_mouse_down[4];

	float					z_master_music_tempo;
	float					z_master_music_volume;
	MusicX*					z_current_music;

	int 					z_memory_used_current; // graphics memory use record (in bytes)
	int 					z_memory_used_max;
	std::vector<ImageX*>	z_queued_image_loads;
	std::vector<ImageX*>	z_copyable_images_loaded;
	std::vector<ImageX*>	z_copying_images;
	int						z_lastPercent; // last percent done loading

	// vital DirectX field variables:

	// DirectDraw 7.0		
	LPDIRECTDRAW7			z_dd_main;
	LPDIRECTDRAWSURFACE7	z_dd_front_buf;
	LPDIRECTDRAWSURFACE7	z_dd_back_buf;
	LPDIRECTDRAWSURFACE7	z_last_texture_surface; // to remember last texture to avoid excess calling of SetTexture ()
	int 					z_dd_was_lost;
	HRESULT 				z_dd_last_error;
	bool					z_win_no_acc_err;

	// Direct3D 7.0
	LPDIRECT3D7 			z_d3d_main;	 // D3D main object
	LPDIRECT3DDEVICE7		z_d3d_device;  // D3D rendering device
	LPDIRECTDRAWSURFACE7	z_d3d_zbuffer; // Z-buffer used for per-pixel depth checking of 3D scene drawing
	bool					z_d3d_zbuffer_enabled;
	bool					z_d3d_zbuffer_wasEnabled;
	int						z_d3d_max_tex_width, z_d3d_max_tex_height; // maximum texture size allowed, lower for decreased resolution
	LPDIRECT3DVERTEXBUFFER7 z_d3d_vbuffer; // Vertex buffer used to batch primitive drawing calls
	FVFTLVERTEX*			z_d3d_vbuffermem;
	int						z_d3d_vbufferfilled; // number of vertices currently in vertex buffer
	int						z_d3d_vbuffermaxverts; // maximum vertices the vertex buffer can hold
	bool					z_d3d_vbufferlocked;
	bool					z_d3d_vbufferenabled;

	// DirectInput 7.0		
	LPDIRECTINPUT			z_di_main;
	LPDIRECTINPUTDEVICE 	z_di_keyboard;
	LPDIRECTINPUTDEVICE 	z_di_mouse;
	unsigned char			z_di_keystate[256];
	unsigned char			z_di_keypressed[256];
	LPDIMOUSESTATE			z_di_mousestate;
	BYTE					z_di_mousepressed[4];
	int 					z_di_mousestop;
	int 					z_di_statMouseX, z_di_statMouseY;

	// DirectSound 3.1
	LPDIRECTSOUND			z_ds_main;
	std::vector<SoundX*>	z_sound_list;

	// DirectShow (see in MusicX)
	bool					z_mp3_support;
	bool					z_mp3_support_tested;

	// DirectMusic
	IDirectMusicPerformance* z_dm_performance;
	IDirectMusicLoader*		 z_dm_loader;
	MusicX*					 z_dm_musicx [256];
	int 					 z_dm_musicx_index;

	bool					z_sound_support;
	bool					z_sound_support_tested;


friend class ImageX;
friend class ImageExt;
friend class SoundX;
friend class MusicX;
friend class MovieX;
};

extern void GameInit ();
extern void GameCleanup ();
extern void GameRun (); // this function should update the game state but NOT draw that state
extern void GameDraw (); // this function should draw the game state and do as little else as possible

#endif